# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM Heat Elmer writer"
__author__ = "Markus Hovorka, Bernd Hahnebach, Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import itertools
import FreeCAD

from .. import sifio
from .. import writer as general_writer
from femtools import membertools
from femmesh import meshtools
from . import heat


class Heatwriter:

    def __init__(self, writer, solver):
        self.write = writer
        self.solver = solver

    def getHeatSolver(self, equation):
        # check if we need to update the equation
        self._updateHeatSolver(equation)
        # output the equation parameters
        s = self.write.createNonlinearSolver(equation)
        s["Equation"] = equation.Name
        s["Procedure"] = sifio.FileAttr("HeatSolve/HeatSolver")
        s["Bubbles"] = equation.Bubbles
        s["Exec Solver"] = "Always"
        s["Optimize Bandwidth"] = True
        s["Stabilize"] = equation.Stabilize
        s["Variable"] = self.write.getUniqueVarName("Temperature")
        return s

    def handleHeatConstants(self):
        self.write.constant(
            "Stefan Boltzmann",
            self.write.convert(self.write.constsdef["StefanBoltzmann"], "M/(O^4*T^3)"),
        )

    def handleHeatEquation(self, bodies, equation):
        for b in bodies:
            if equation.Convection != "None":
                self.write.equation(b, "Convection", equation.Convection)
            if equation.PhaseChangeModel != "None":
                self.write.equation(b, "Phase Change Model", equation.PhaseChangeModel)

    def _updateHeatSolver(self, equation):
        # updates older Heat equations
        if not hasattr(equation, "Convection"):
            equation.addProperty(
                "App::PropertyEnumeration",
                "Convection",
                "Equation",
                "Type of convection to be used",
                locked=True,
            )
            equation.Convection = heat.CONVECTION_TYPE
            equation.Convection = "None"
        if not hasattr(equation, "PhaseChangeModel"):
            equation.addProperty(
                "App::PropertyEnumeration",
                "PhaseChangeModel",
                "Equation",
                "Model for phase change",
                locked=True,
            )
            equation.PhaseChangeModel = heat.PHASE_CHANGE_MODEL
            equation.PhaseChangeModel = "None"

    def handleHeatBndConditions(self):
        i = -1
        for obj in self.write.getMember("Fem::ConstraintTemperature"):
            i = i + 1
            femobjects = membertools.get_several_member(
                self.write.analysis, "Fem::ConstraintTemperature"
            )
            femobjects[i]["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(
                self.write.getSingleMember("Fem::FemMeshObject").FemMesh, femobjects[i]
            )
            NumberOfNodes = len(femobjects[i]["Nodes"])
            if obj.References:
                for name in obj.References[0][1]:
                    if obj.ConstraintType == "Temperature":
                        temperature = float(obj.Temperature.getValueAs("K"))
                        self.write.boundary(name, "Temperature", temperature)
                    elif obj.ConstraintType == "CFlux":
                        flux = float(obj.CFlux.getValueAs("W"))
                        # CFLUX is the flux per mesh node
                        flux = flux / NumberOfNodes
                        self.write.boundary(name, "Temperature Load", flux)
                self.write.handled(obj)
        for obj in self.write.getMember("Fem::ConstraintHeatflux"):
            if obj.References:
                for name in obj.References[0][1]:
                    if obj.ConstraintType == "Convection":
                        film = obj.FilmCoef.getValueAs("W/(m^2*K)").Value
                        temp = obj.AmbientTemp.getValueAs("K").Value
                        self.write.boundary(name, "Heat Transfer Coefficient", film)
                        self.write.boundary(name, "External Temperature", temp)
                    elif obj.ConstraintType == "DFlux":
                        flux = obj.DFlux.getValueAs("W/m^2").Value
                        self.write.boundary(name, "Heat Flux BC", True)
                        self.write.boundary(name, "Heat Flux", flux)
                self.write.handled(obj)

    def handleHeatInitial(self, bodies):
        tempObj = self.write.getSingleMember("Fem::ConstraintInitialTemperature")
        if tempObj is not None:
            refTemp = float(tempObj.initialTemperature.getValueAs("K"))
            for name in bodies:
                self.write.initial(name, "Temperature", refTemp)
            self.write.handled(tempObj)

    def _outputHeatBodyForce(self, obj, name):
        if obj.Mode == "Dissipation Rate":
            heatSource = obj.DissipationRate.getValueAs("W/kg").Value

        elif obj.Mode == "Total Power":
            ref = obj.References[0]
            ref_feat = ref[0]
            ref_sub_obj = ref[1][0]
            density = None
            for mat in self.write.getMember("App::MaterialObject"):
                mat_ref = [
                    *itertools.chain(*[itertools.product([i[0]], i[1]) for i in mat.References])
                ]
                if (ref_feat, ref_sub_obj) in mat_ref:
                    density = FreeCAD.Units.Quantity(mat.Material["Density"])
                    break

            if not density:
                # search material without references
                for mat in self.write.getMember("App::MaterialObject"):
                    if not mat.References:
                        density = FreeCAD.Units.Quantity(mat.Material["Density"])
                        break
            volume = ref_feat.getSubObject(ref_sub_obj).Volume
            heatSource = (
                (obj.TotalPower / (density * FreeCAD.Units.Quantity(volume, "mm^3")))
                .getValueAs("W/kg")
                .Value
            )

        if heatSource == 0.0:
            # a zero heat would break Elmer (division by zero)
            raise general_writer.WriteError("The body heat source must not be zero!")
        self.write.bodyForce(name, "Heat Source", heatSource)

    def handleHeatBodyForces(self, bodies):
        bodyHeats = self.write.getMember("Fem::ConstraintBodyHeatSource")
        for obj in bodyHeats:
            if obj.References:
                for name in obj.References[0][1]:
                    self._outputHeatBodyForce(obj, name)
                self.write.handled(obj)
            else:
                # if there is only one body heat without a reference
                # add it to all bodies
                if len(bodyHeats) == 1:
                    for name in bodies:
                        self._outputHeatBodyForce(obj, name)
                else:
                    raise general_writer.WriteError(
                        "Several body heat constraints found without reference to a body.\n"
                        "Please set a body for each body heat constraint."
                    )
            self.write.handled(obj)

    def handleHeatMaterial(self, bodies):
        tempObj = self.write.getSingleMember("Fem::ConstraintInitialTemperature")
        if tempObj is not None:
            refTemp = float(tempObj.initialTemperature.getValueAs("K"))
            for name in bodies:
                self.write.material(name, "Reference Temperature", refTemp)
        for obj in self.write.getMember("App::MaterialObject"):
            m = obj.Material
            refs = obj.References[0][1] if obj.References else self.write.getAllBodies()
            for name in (n for n in refs if n in bodies):
                if "Density" not in m:
                    raise general_writer.WriteError(
                        "Used material does not specify the necessary 'Density'."
                    )
                self.write.material(name, "Name", m["Name"])
                self.write.material(name, "Density", self.write.getDensity(m))
                if "ThermalConductivity" not in m:
                    raise general_writer.WriteError(
                        "Used material does not specify the necessary 'Thermal Conductivity'."
                    )
                self.write.material(
                    name,
                    "Heat Conductivity",
                    self.write.convert(m["ThermalConductivity"], "M*L/(T^3*O)"),
                )
                if "SpecificHeat" not in m:
                    raise general_writer.WriteError(
                        "Used material does not specify the necessary 'Specific Heat'."
                    )
                self.write.material(
                    name, "Heat Capacity", self.write.convert(m["SpecificHeat"], "L^2/(T^2*O)")
                )


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2022 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import FreeCADGui
import Path
import Path.Dressup.DogboneII as DogboneII
import Path.Dressup.Utils as PathDressup
import PathScripts.PathUtils as PathUtils

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class Marker(object):
    def __init__(self, pt, r, h, ena):
        if Path.Geom.isRoughly(h, 0):
            h = 0.1
        self.pt = pt
        self.r = r
        self.h = h
        self.ena = ena
        self.sep = coin.SoSeparator()
        self.pos = coin.SoTranslation()
        self.pos.translation = (pt.x, pt.y, pt.z + h / 2)
        self.rot = coin.SoRotationXYZ()
        self.rot.axis = self.rot.X
        self.rot.angle = DogboneII.PI / 2
        self.cyl = coin.SoCylinder()
        self.cyl.radius = r
        self.cyl.height = h
        # self.cyl.removePart(self.cyl.TOP)
        # self.cyl.removePart(self.cyl.BOTTOM)
        self.material = coin.SoMaterial()
        self.sep.addChild(self.pos)
        self.sep.addChild(self.rot)
        self.sep.addChild(self.material)
        self.sep.addChild(self.cyl)
        self.lowlight()

    def setSelected(self, selected):
        if selected:
            self.highlight()
        else:
            self.lowlight()

    def highlight(self):
        self.material.diffuseColor = self.color(1)
        self.material.transparency = 0.75

    def lowlight(self):
        self.material.diffuseColor = self.color(0)
        self.material.transparency = 0.90

    def _colorEnabled(self, id):
        if id == 1:
            return coin.SbColor(0.0, 0.9, 0.0)
        return coin.SbColor(0.0, 0.9, 0.0)

    def _colorDisabled(self, id):
        if id == 1:
            return coin.SbColor(0.9, 0.0, 0.0)
        return coin.SbColor(0.9, 0.0, 0.0)

    def color(self, id):
        if self.ena:
            return self._colorEnabled(id)
        return self._colorDisabled(id)


class TaskPanel(object):
    DataIds = QtCore.Qt.ItemDataRole.UserRole
    DataState = QtCore.Qt.ItemDataRole.UserRole + 1

    def __init__(self, viewProvider, obj):
        self.viewProvider = viewProvider
        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/DogboneEdit.ui")
        self.s = None
        FreeCAD.ActiveDocument.openTransaction("Edit Dogbone Dress-up")
        # self.height = 10 ???
        self.markers = []

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.Selection.removeObserver(self.s)
        self.cleanup()

    def accept(self):
        self.getFields()
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.Selection.removeObserver(self.s)
        FreeCAD.ActiveDocument.recompute()
        self.cleanup()

    def cleanup(self):
        self.viewProvider.showMarkers(False)
        for m in self.markers:
            self.viewProvider.switch.removeChild(m.sep)
        self.markers = []

    def getFields(self):
        self.obj.Style = str(self.form.styleCombo.currentText())
        self.obj.Side = str(self.form.sideCombo.currentText())
        self.obj.Incision = str(self.form.incisionCombo.currentText())
        self.obj.Custom = self.form.custom.value()
        blacklist = []
        for i in range(0, self.form.bones.count()):
            item = self.form.bones.item(i)
            if item.checkState() == QtCore.Qt.CheckState.Unchecked:
                blacklist.extend(item.data(self.DataIds))
        self.obj.BoneBlacklist = sorted(blacklist)
        self.obj.Proxy.execute(self.obj)

    def updateBoneList(self):
        itemList = []
        for state in self.obj.Proxy.boneStates(self.obj):
            pos = state.position()
            ids = ",".join([str(nr) for nr in state.boneIDs()])
            lbl = f"({pos.x:.2f}, {pos.y:.2f}): {ids}"
            item = QtGui.QListWidgetItem(lbl)
            if state.isEnabled():
                item.setCheckState(QtCore.Qt.CheckState.Checked)
            else:
                item.setCheckState(QtCore.Qt.CheckState.Unchecked)
            flags = QtCore.Qt.ItemFlag.ItemIsSelectable
            flags |= QtCore.Qt.ItemFlag.ItemIsEnabled
            flags |= QtCore.Qt.ItemFlag.ItemIsUserCheckable
            item.setFlags(flags)
            item.setData(self.DataIds, state.boneIDs())
            item.setData(self.DataState, state)
            itemList.append(item)

        markers = []
        self.form.bones.clear()
        for item in sorted(itemList, key=lambda item: item.data(self.DataState).boneIDs()[0]):
            self.form.bones.addItem(item)
            state = item.data(self.DataState)
            loc = state.boneTip()
            r = self.obj.Proxy.toolRadius(self.obj)
            zs = state.zLevels()
            markers.append(
                Marker(
                    FreeCAD.Vector(loc.x, loc.y, min(zs)),
                    r,
                    max(1, max(zs) - min(zs)),
                    state.isEnabled(),
                )
            )
        for m in self.markers:
            self.viewProvider.switch.removeChild(m.sep)
        for m in markers:
            self.viewProvider.switch.addChild(m.sep)
        self.markers = markers

    def updateUI(self):
        customSelected = self.obj.Incision == DogboneII.Incision.Custom
        self.form.custom.setEnabled(customSelected)
        self.form.customLabel.setEnabled(customSelected)
        self.updateBoneList()

    def updateModel(self):
        self.getFields()
        self.updateUI()
        FreeCAD.ActiveDocument.recompute()

    def setupCombo(self, combo, text, items):
        if items and len(items) > 0:
            for i in range(combo.count(), -1, -1):
                combo.removeItem(i)
            combo.addItems(items)
        index = combo.findText(text, QtCore.Qt.MatchFixedString)
        if index >= 0:
            combo.setCurrentIndex(index)

    def setFields(self):
        self.setupCombo(self.form.styleCombo, self.obj.Style, DogboneII.Style.All)
        self.setupCombo(self.form.sideCombo, self.obj.Side, DogboneII.Side.All)
        self.setupCombo(self.form.incisionCombo, self.obj.Incision, DogboneII.Incision.All)
        self.form.custom.setMinimum(0.0)
        self.form.custom.setDecimals(3)
        self.form.custom.setValue(self.obj.Custom)
        self.updateUI()

    def open(self):
        self.s = SelObserver()
        # install the function mode resident
        FreeCADGui.Selection.addObserver(self.s)

    def setupUi(self):
        self.setFields()
        # now that the form is filled, setup the signal handlers
        self.form.styleCombo.currentIndexChanged.connect(self.updateModel)
        self.form.sideCombo.currentIndexChanged.connect(self.updateModel)
        self.form.incisionCombo.currentIndexChanged.connect(self.updateModel)
        self.form.custom.valueChanged.connect(self.updateModel)
        self.form.bones.itemChanged.connect(self.updateModel)
        self.form.bones.itemSelectionChanged.connect(self.updateMarkers)

        self.viewProvider.showMarkers(True)

    def updateMarkers(self):
        index = self.form.bones.currentRow()
        for i, m in enumerate(self.markers):
            m.setSelected(i == index)


class SelObserver(object):
    def __init__(self):
        import Path.Op.Gui.Selection as PST

        PST.eselect()

    def __del__(self):
        import Path.Op.Gui.Selection as PST

        PST.clear()

    def addSelection(self, doc, obj, sub, pnt):
        FreeCADGui.doCommand("Gui.Selection.addSelection(FreeCAD.ActiveDocument." + obj + ")")
        FreeCADGui.updateGui()


class ViewProviderDressup(object):
    def __init__(self, vobj):
        self.vobj = vobj
        self.obj = None

    def attach(self, vobj):
        self.obj = vobj.Object
        if self.obj and self.obj.Base:
            for i in self.obj.Base.InList:
                if hasattr(i, "Group"):
                    group = i.Group
                    for g in group:
                        if g.Name == self.obj.Base.Name:
                            group.remove(g)
                    i.Group = group
            # FreeCADGui.ActiveDocument.getObject(obj.Base.Name).Visibility = False
        self.switch = coin.SoSwitch()
        vobj.RootNode.addChild(self.switch)

    def showMarkers(self, on):
        sw = coin.SO_SWITCH_ALL if on else coin.SO_SWITCH_NONE
        self.switch.whichChild = sw

    def claimChildren(self):
        return [self.obj.Base]

    def setEdit(self, vobj, mode=0):
        FreeCADGui.Control.closeDialog()
        panel = TaskPanel(self, vobj.Object)
        FreeCADGui.Control.showDialog(panel)
        panel.setupUi()
        return True

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onDelete(self, arg1=None, arg2=None):
        """this makes sure that the base operation is added back to the project and visible"""
        if arg1.Object and arg1.Object.Base:
            FreeCADGui.ActiveDocument.getObject(arg1.Object.Base.Name).Visibility = True
            job = PathUtils.findParentJob(arg1.Object)
            if job:
                job.Proxy.addOperation(arg1.Object.Base, arg1.Object)
            arg1.Object.Base = None
        return True

    def getIcon(self):
        if getattr(PathDressup.baseOp(self.obj), "Active", True):
            return ":/icons/CAM_Dressup.svg"
        else:
            return ":/icons/CAM_OpActive.svg"


def Create(base, name="DressupDogbone"):
    """
    Create(obj, name='DressupDogbone')… dresses the given Path.Op.Profile object with dogbones.
    """
    obj = DogboneII.Create(base, name)
    job = PathUtils.findParentJob(base)
    job.Proxy.addOperation(obj, base)

    if FreeCAD.GuiUp:
        obj.ViewObject.Proxy = ViewProviderDressup(obj.ViewObject)
        obj.Base.ViewObject.Visibility = False

    return obj


class CommandDressupDogboneII(object):
    def GetResources(self):
        return {
            "Pixmap": "CAM_Dressup",
            "MenuText": QT_TRANSLATE_NOOP("CAM_DressupDogbone", "Dogbone"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_DressupDogbone",
                "Creates a dogbone dress-up object from a selected toolpath",
            ),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):

        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            FreeCAD.Console.PrintError(
                translate("CAM_DressupDogbone", "Select one toolpath object") + "\n"
            )
            return
        baseObject = selection[0]
        if not baseObject.isDerivedFrom("Path::Feature"):
            FreeCAD.Console.PrintError(
                translate("CAM_DressupDogbone", "The selected object is not a toolpath") + "\n"
            )
            return

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction("Create Dogbone Dress-up")
        FreeCADGui.addModule("Path.Dressup.Gui.DogboneII")
        FreeCADGui.doCommand(
            "Path.Dressup.Gui.DogboneII.Create(FreeCAD.ActiveDocument.%s)" % baseObject.Name
        )
        # FreeCAD.ActiveDocument.commitTransaction()  # Final `commitTransaction()` called via TaskPanel.accept()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtGui
    from pivy import coin

    FreeCADGui.addCommand("CAM_DressupDogbone", CommandDressupDogboneII())

FreeCAD.Console.PrintLog("Loading DressupDogboneII… done\n")



# ***************************************************************************
# *   Copyright (c) 2022 Wanderer Fan <wandererfan@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides utility functions for TD Tools."""

import FreeCAD as App
import FreeCADGui as Gui
import Part
from PySide.QtCore import QT_TRANSLATE_NOOP
from PySide import QtGui
from PySide.QtGui import QMessageBox

def havePage():
    objs = App.ActiveDocument.Objects
    for o in objs:
        if o.isDerivedFrom("TechDraw::DrawPage"):
            return True
    return False

def haveView():
    objs = App.ActiveDocument.Objects
    for o in objs:
        if o.isDerivedFrom("TechDraw::DrawView"):
            return True
    return False

def displayMessage(title,message):
    '''
    displayMessage('Title','Messagetext')
    '''
    msgBox = QtGui.QMessageBox()
    msgBox.setText(message)
    msgBox.setWindowTitle(title)
    msgBox.exec_()


def getSelView(nSel=0):
    '''
    view = getSelView()
    nSel=0 ... number of selected view, 0 = first selected
    Return selected view, otherwise return False
    '''
    if not Gui.Selection.getSelection():
        displayMessage('TechDraw_Utils','No view selected')
    else:
        view = Gui.Selection.getSelection()[nSel]
        return view

def getSelVertexes(nVertex=1, nSel=0):
    '''
    vertexes = getSelVertexes(nVertex)
    nVertex=1 ... min. number of selected vertexes
    nSel=0 ... number of selected view, 0 = first selected
    Return a list of selected vertexes if at least nVertex vertexes are selected, otherwise return False
    '''
    if getSelView(nSel):
        view = getSelView(nSel)
    else:
        return False
    if not Gui.Selection.getSelectionEx():
        displayMessage('TechDraw_Utils',
                        QT_TRANSLATE_NOOP('TechDraw_Utils','No vertex selected'))
        return False
    objectList = Gui.Selection.getSelectionEx()[nSel].SubElementNames

    vertexes = []
    for objectString in objectList:
        if objectString[0:6] == 'Vertex':
            vertexes.append(view.getVertexBySelection(objectString))

    if (len(vertexes) < nVertex):
        displayMessage('TechDraw_Utils',
                        QT_TRANSLATE_NOOP('TechDraw_Utils','Select at least ')+
                        str(nVertex)+
                        QT_TRANSLATE_NOOP('TechDraw_Utils',' vertexes'))
        return False
    else:
        return vertexes

def getSelEdges(nEdge=1, nSel=0):
    '''
    edges = getSelEdges(nEdge)
    nEdge=1 ... min. number of selected edges
    nSel=0 ... number of selected view, 0 = first selected
    Return a list of selected edges if at least nedge edges are selected, otherwise return False
    '''
    if getSelView(nSel):
        view = getSelView(nSel)
    else:
        return False
    if not Gui.Selection.getSelectionEx():
        displayMessage('TechDraw_Utils',
                        QT_TRANSLATE_NOOP('TechDraw_Utils','No edge selected'))
        return False
    objectList = Gui.Selection.getSelectionEx()[nSel].SubElementNames

    edges = []
    for objectString in objectList:
        if objectString[0:4] == 'Edge':
            edges.append(view.getEdgeBySelection(objectString))

    if (len(edges) < nEdge):
        displayMessage('TechDraw_Utils',
                        QT_TRANSLATE_NOOP('TechDraw_Utils','Select at least ')+
                        str(nEdge)+
                        QT_TRANSLATE_NOOP('TechDraw_Utils',' edges'))
        return False
    else:
        return edges

def getCoordinateVectors(view):
    '''
    (px,py,pz) = getCoordinateVectors(view)
    view ... selected view
    (px,py,pz) ... returned tuple of vectors (App.Vector)
    calculate projected vectors of x-, y- and z-axis
    '''
    diagonal = view.Direction
    xDirection = view.XDirection
    origin = App.Vector()
    xAxisProj = App.Vector(1,0,0).projectToPlane(origin,diagonal)
    yAxisProj = App.Vector(0,1,0).projectToPlane(origin,diagonal)
    zAxisProj = App.Vector(0,0,1).projectToPlane(origin,diagonal)
    # create wire polygon and roll back in 3D
    wire3D = Part.makePolygon([xAxisProj,yAxisProj,zAxisProj,xDirection,xAxisProj])
    rotation3D = App.Rotation(diagonal,App.Vector(0,0,1))
    wire3D.Placement.Rotation = rotation3D
    px = wire3D.Edges[0].Vertexes[0].Point
    py = wire3D.Edges[1].Vertexes[0].Point
    pz = wire3D.Edges[2].Vertexes[0].Point
    pDir = wire3D.Edges[3].Vertexes[0].Point
    pDir.z = 0.0
    # rotate inside drawing plane
    wire2D = Part.makePolygon([px,py,pz,px])
    rotation2D = App.Rotation(pDir,App.Vector(1,0,0))
    wire2D.Placement.Rotation = rotation2D
    px = wire2D.Edges[0].Vertexes[0].Point
    py = wire2D.Edges[1].Vertexes[0].Point
    pz = wire2D.Edges[2].Vertexes[0].Point
    return (px,py,pz)
    

# +/- the same as getSelVertexes, but returns the vertex name instead of the vertex object.
def getSelVertexNames(nVertex=1, nSel=0):
    '''
    vertexNames = getSelVertexNames(nVertex)
    nVertex=1 ... min. number of selected vertexes
    nSel=0 ... number of selected view, 0 = first selected
    Return a list of the names of the selected vertexes if at least nVertex vertexes are selected, otherwise return False
    '''
    view = getSelView(nSel)
    if not view:
        return False
        
    if not Gui.Selection.getSelectionEx():
        displayMessage('TechDraw_Utils',
                        QT_TRANSLATE_NOOP('TechDraw_Utils','No vertex selected'))
        return False
    objectList = Gui.Selection.getSelectionEx()[nSel].SubElementNames

    vertexNames = []
    for objectString in objectList:
        if objectString[0:6] == 'Vertex':
            vertexNames.append(objectString)

    if (len(vertexNames) < nVertex):
        displayMessage('TechDraw_Utils',
                        QT_TRANSLATE_NOOP('TechDraw_Utils','Select at least ')+
                        str(nVertex)+
                        QT_TRANSLATE_NOOP('TechDraw_Utils',' vertexes'))
        return False

    return vertexNames

# +/- the same as getSelEdges, but returns the edge name instead of the vertex object.
def getSelEdgeNames(nEdge=1, nSel=0):
    '''
    edges = getSelEdgeNames(nEdge)
    nEdge=1 ... min. number of selected edges
    nSel=0 ... number of selected view, 0 = first selected
    Return a list of names for selected edges if at least nedge edges are selected, otherwise return False
    '''
    view = getSelView(nSel)
    if not view:
        return False
        
    if not Gui.Selection.getSelectionEx():
        displayMessage('TechDraw_Utils',
                        QT_TRANSLATE_NOOP('TechDraw_Utils','No edge selected'))
        return False
    objectList = Gui.Selection.getSelectionEx()[nSel].SubElementNames

    edgeNames = []
    for objectString in objectList:
        if objectString[0:4] == 'Edge':
            edgeNames.append(objectString)

    if (len(edgeNames) < nEdge):
        displayMessage('TechDraw_Utils',
                        QT_TRANSLATE_NOOP('TechDraw_Utils','Select at least ')+
                        str(nEdge)+
                        QT_TRANSLATE_NOOP('TechDraw_Utils',' edges'))
        return False
    else:
        return edgeNames



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""The MacroCacheCreator is an independent script run server-side to generate a cache of
the macros and their metadata. Supports both git-based and wiki-based macros."""

import hashlib
import json
import os.path
import re
import urllib.parse

import requests
from typing import Dict
import zipfile

from addonmanager_macro import Macro
from AddonCatalogCacheCreator import CacheWriter  # Borrow the git utility method from this class

GIT_MACROS_URL = "https://github.com/FreeCAD/FreeCAD-macros.git"
GIT_MACROS_BRANCH = "master"
GIT_MACROS_CLONE_NAME = "FreeCAD-macros"

WIKI_MACROS_URL = "https://wiki.freecad.org/Macros_recipes"

# Several of these are really just artifacts from the wiki page, not real macros at all.
MACROS_REJECT_LIST = [
    "BOLTS",
    "WorkFeatures",
    "how to install",
    "documentation",
    "PartsLibrary",
    "FCGear",
]

headers = {"User-Agent": b"FreeCAD AddonManager/1.0"}


class MacroCatalog:
    """A catalog of macros."""

    def __init__(self):
        self.macros: Dict[str, Macro] = {}
        self.macro_errors = {}
        self.macro_stats = {
            "macros_on_wiki": 0,
            "macros_on_git": 0,
            "duplicated_macros": 0,
            "errors": 0,
        }

    def fetch_macros(self):
        print("Retrieving macros from git...")
        self.retrieve_macros_from_git()
        print("Retrieving macros from wiki...")
        self.retrieve_macros_from_wiki()
        print("Downloading icons...")
        for macro in self.macros.values():
            try:
                self.get_icon(macro)
            except RuntimeError as e:
                self.macro_errors[macro.name] = str(e)
        self.macro_stats["errors"] = len(self.macro_errors)

    def create_cache(self) -> str:
        """Create a cache from the macros in this catalog"""
        cache_dict = {}
        for macro in self.macros.values():
            cache_dict[macro.name] = macro.to_cache()
        return json.dumps(cache_dict, indent=4)

    def retrieve_macros_from_git(self):
        """Retrieve macros from GIT_MACROS_URL"""

        try:
            writer = CacheWriter()
            writer.clone_or_update(GIT_MACROS_CLONE_NAME, GIT_MACROS_URL, GIT_MACROS_BRANCH)
        except RuntimeError as e:
            print(f"Failed to clone git macros from {GIT_MACROS_URL}: {e}")
            return

        for dirpath, _, filenames in os.walk(os.path.join(os.getcwd(), GIT_MACROS_CLONE_NAME)):
            if ".git" in dirpath:
                continue
            for filename in filenames:
                if filename.lower().endswith(".fcmacro"):
                    self.macro_stats["macros_on_git"] += 1
                    self.add_git_macro_to_cache(dirpath, filename)

    def add_git_macro_to_cache(self, dirpath: str, filename: str):
        macro = Macro(filename[:-8])  # Remove ".FCMacro".
        if macro.name in self.macros:
            print(f"Ignoring second macro named {macro.name} (found on git)\n")
            return
        macro.on_git = True
        absolute_path_to_fcmacro = os.path.join(dirpath, filename)
        macro.fill_details_from_file(absolute_path_to_fcmacro)
        macro.src_filename = os.path.relpath(absolute_path_to_fcmacro, os.getcwd())
        self.macros[macro.name] = macro

    def retrieve_macros_from_wiki(self):
        """Retrieve macros from the wiki

        Read the wiki and add a cache entry for each found macro.
        Reads only the page https://wiki.freecad.org/Macros_recipes
        """

        try:
            p = requests.get(WIKI_MACROS_URL, headers=headers, timeout=10.0)
        except requests.exceptions.RequestException as e:
            message = f"Failed to fetch {WIKI_MACROS_URL}: {e}"
            self.macro_errors["retrieve_macros_from_wiki"] = message
            return
        if not p.status_code == 200:
            print(f"Failed to fetch {WIKI_MACROS_URL}, response code was {p.status_code}")
            return

        macros = re.findall(r'title="(Macro.*?)"', p.text)
        macros = [mac for mac in macros if "translated" not in mac]
        for _, wiki_page_name in enumerate(macros):
            macro_name = wiki_page_name[6:]  # Remove "Macro ".
            macro_name = macro_name.replace("&amp;", "&")
            if not macro_name:
                continue
            if (macro_name not in MACROS_REJECT_LIST) and ("recipes" not in macro_name.lower()):
                self.macro_stats["macros_on_wiki"] += 1
                self.add_wiki_macro_to_cache(macro_name)

    def add_wiki_macro_to_cache(self, macro_name):
        macro = Macro(macro_name)
        if macro.name in self.macros:
            self.macro_stats["duplicated_macros"] += 1
            print(f"Ignoring duplicate of '{macro.name}' (using git repo copy instead of wiki)")
            return
        macro.on_wiki = True
        macro.parsed = False
        self.macros[macro.name] = macro
        wiki_page_name = macro.name.replace(" ", "_")
        wiki_page_name = wiki_page_name.replace("&", "%26")
        wiki_page_name = wiki_page_name.replace("+", "%2B")
        url = "https://wiki.freecad.org/Macro_" + wiki_page_name
        macro.fill_details_from_wiki(url)

    @staticmethod
    def get_icon(macro: Macro):
        """Downloads the macro's icon from whatever source is specified and stores its binary
        contents in self.icon_data"""
        if macro.icon.startswith("http://") or macro.icon.startswith("https://"):
            if "freecadweb" in macro.icon:
                macro.icon = macro.icon.replace("freecadweb", "freecad")
            parsed_url = urllib.parse.urlparse(macro.icon)
            try:
                p = requests.get(macro.icon, headers=headers, timeout=10.0)
            except requests.exceptions.RequestException as e:
                message = f"Failed to get data from icon URL {macro.icon}: {e}"
                self.macro_errors[macro.name] = message
                macro.icon = ""
                return
            if p.status_code == 200:
                _, _, filename = parsed_url.path.rpartition("/")
                base, _, extension = filename.rpartition(".")
                if base.lower().startswith("file:"):
                    print(
                        f"Cannot use specified icon for {macro.name}, {macro.icon} "
                        "is not a direct download link"
                    )
                    macro.icon = ""
                    return
                macro.icon_data = p.content
                macro.icon_extension = extension
            else:
                print(
                    f"MACRO DEVELOPER WARNING: failed to download icon from {macro.icon}"
                    f" for macro {macro.name}. Status code returned: {p.status_code}\n"
                )
                macro.icon = ""
        elif macro.on_git:
            relative_path_to_macro_directory = os.path.dirname(macro.src_filename)
            if "/" in macro.icon:
                relative_path_to_icon = macro.icon.replace("/", os.path.sep)
            else:
                relative_path_to_icon = macro.icon
            local_icon = os.path.join(
                os.getcwd(), relative_path_to_macro_directory, relative_path_to_icon
            )
            if os.path.isfile(local_icon):
                with open(local_icon, "rb") as icon_file:
                    macro.icon_data = icon_file.read()
                    macro.icon_extension = relative_path_to_icon.rpartition(".")[-1]


if __name__ == "__main__":
    catalog = MacroCatalog()
    catalog.fetch_macros()
    cache = catalog.create_cache()

    with zipfile.ZipFile(
        os.path.join(os.getcwd(), "macro_cache.zip"), "w", zipfile.ZIP_DEFLATED
    ) as zipf:
        zipf.writestr("macro_cache.json", cache)

    # Also generate the sha256 hash of the zip file and store it
    with open("macro_cache.zip", "rb") as cache_file:
        cache_file_content = cache_file.read()
    sha256 = hashlib.sha256(cache_file_content).hexdigest()
    with open("macro_cache.zip.sha256", "w", encoding="utf-8") as hash_file:
        hash_file.write(sha256)

    # Finally, write out the errors and stats as JSON data:
    with open(os.path.join(os.getcwd(), "macro_errors.json"), "w", encoding="utf-8") as f:
        json.dump(catalog.macro_errors, f, indent="  ")
    with open(os.path.join(os.getcwd(), "macro_stats.json"), "w", encoding="utf-8") as f:
        json.dump(catalog.macro_stats, f, indent="  ")

    print("Cache written to macro_cache.zip and macro_cache.zip.sha256")



# /***************************************************************************
# *   Copyright (c) 2019 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

from Show.SceneDetail import SceneDetail


class Pickability(SceneDetail):
    """Pickability(object, pickstyle = None):Plugin for TempoVis for altering pick style
    of objects (i.e., selectability).
    pickstyle may be:
        PS_REGULAR = 0 # selectable
        PS_BOUNDBOX = 1 # selectable, but faster hit testing using bounding box
        PS_UNPICKABLE = 2 # not selectable and not obstructing."""

    class_id = "SDPickability"
    propname = ""
    objname = ""

    def __init__(self, object, pickstyle=None):
        self.objname = object.Name
        self.doc = object.Document
        self.key = self.objname
        if pickstyle is not None:
            self.data = pickstyle

    def scene_value(self):
        return getPickStyle(self.doc.getObject(self.objname).ViewObject)

    def apply_data(self, val):
        setPickStyle(self.doc.getObject(self.objname).ViewObject, val)


PS_REGULAR = 0
PS_BOUNDBOX = 1
PS_UNPICKABLE = 2


def getPickStyleNode(viewprovider, make_if_missing=True):
    from pivy import coin

    sa = coin.SoSearchAction()
    sa.setType(coin.SoPickStyle.getClassTypeId())
    sa.traverse(viewprovider.RootNode)
    if sa.isFound() and sa.getPath().getLength() == 1:
        return sa.getPath().getTail()
    else:
        if not make_if_missing:
            return None
        pick_style = coin.SoPickStyle()
        pick_style.style.setValue(coin.SoPickStyle.SHAPE)
        viewprovider.RootNode.insertChild(pick_style, 0)
        return pick_style


def getPickStyle(viewprovider):
    ps = getPickStyleNode(viewprovider, make_if_missing=False)
    if ps is not None:
        return ps.style.getValue()
    else:
        return PS_REGULAR


def setPickStyle(viewprovider, pickstyle):
    ps = getPickStyleNode(viewprovider, make_if_missing=pickstyle != 0)  # coin.SoPickStyle.SHAPE
    if ps is not None:
        return ps.style.setValue(pickstyle)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from Path.Post.Processor import PostProcessorFactory  # PostProcessor,
from PySide import QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import Path
import Path.Base.SetupSheet as PathSetupSheet
import Path.Base.Util as PathUtil
import Path.Main.Stock as PathStock
import Path.Tool.Controller as PathToolController
import json
import time


# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Draft = LazyLoader("Draft", globals(), "Draft")


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class JobTemplate:
    """Attribute and sub element strings for template export/import."""

    Description = "Desc"
    GeometryTolerance = "Tolerance"
    Job = "Job"
    PostProcessor = "Post"
    PostProcessorArgs = "PostArgs"
    PostProcessorOutputFile = "Output"
    Fixtures = "Fixtures"
    OrderOutputBy = "OrderOutputBy"
    SplitOutput = "SplitOutput"
    SetupSheet = "SetupSheet"
    Stock = "Stock"
    # TCs are grouped under Tools in a job, the template refers to them directly though
    ToolController = "ToolController"
    Version = "Version"


def isResourceClone(obj, propLink, resourceName):
    if hasattr(propLink, "PathResource") and (
        resourceName is None or resourceName == propLink.PathResource
    ):
        return True
    return False


def createResourceClone(obj, orig, name, icon):
    clone = Draft.clone(orig)
    clone.Label = "%s-%s" % (name, orig.Label)
    clone.addProperty("App::PropertyString", "PathResource")
    clone.PathResource = name
    if clone.ViewObject:
        import Path.Base.Gui.IconViewProvider

        Path.Base.Gui.IconViewProvider.Attach(clone.ViewObject, icon)
        clone.ViewObject.Visibility = False
        clone.ViewObject.Transparency = 80
    obj.Document.recompute()  # necessary to create the clone shape
    return clone


def createModelResourceClone(obj, orig):
    return createResourceClone(obj, orig, "Model", "BaseGeometry")


class NotificationClass(QtCore.QObject):
    updateTC = QtCore.Signal(object, object)


Notification = NotificationClass()


class ObjectJob:
    def __init__(self, obj, models, templateFile=None):
        self.obj = obj
        self.tooltip = None
        self.tooltipArgs = None
        obj.Proxy = self

        obj.addProperty(
            "App::PropertyFile",
            "PostProcessorOutputFile",
            "Output",
            QT_TRANSLATE_NOOP("App::Property", "The G-code output file for this project"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "PostProcessor",
            "Output",
            QT_TRANSLATE_NOOP("App::Property", "Select the Post Processor"),
        )
        obj.addProperty(
            "App::PropertyString",
            "PostProcessorArgs",
            "Output",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Arguments for the Post Processor (specific to the script)",
            ),
        )
        obj.addProperty(
            "App::PropertyString",
            "LastPostProcessDate",
            "Output",
            QT_TRANSLATE_NOOP("App::Property", "Last Time the Job was post processed"),
        )
        obj.setEditorMode("LastPostProcessDate", 2)  # Hide
        obj.addProperty(
            "App::PropertyString",
            "LastPostProcessOutput",
            "Output",
            QT_TRANSLATE_NOOP("App::Property", "Last Time the Job was post processed"),
        )
        obj.setEditorMode("LastPostProcessOutput", 2)  # Hide

        obj.addProperty(
            "App::PropertyString",
            "Description",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "An optional description for this job"),
        )
        obj.addProperty(
            "App::PropertyString",
            "CycleTime",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Job Cycle Time Estimation"),
        )
        obj.setEditorMode("CycleTime", 1)  # read-only
        obj.addProperty(
            "App::PropertyDistance",
            "GeometryTolerance",
            "Geometry",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "For computing Paths; smaller increases accuracy, but slows down computation",
            ),
        )

        obj.addProperty(
            "App::PropertyLink",
            "Stock",
            "Base",
            QT_TRANSLATE_NOOP("App::Property", "Solid object to be used as stock."),
        )
        obj.addProperty(
            "App::PropertyLink",
            "Operations",
            "Base",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Compound path of all operations in the order they are processed.",
            ),
        )

        obj.addProperty(
            "App::PropertyEnumeration",
            "JobType",
            "Base",
            QT_TRANSLATE_NOOP("App::Property", "Select the Type of Job"),
        )
        obj.setEditorMode("JobType", 2)  # Hide

        obj.addProperty(
            "App::PropertyBool",
            "SplitOutput",
            "Output",
            QT_TRANSLATE_NOOP("App::Property", "Split output into multiple G-code files"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "OrderOutputBy",
            "WCS",
            QT_TRANSLATE_NOOP("App::Property", "If multiple WCS, order the output this way"),
        )
        obj.addProperty(
            "App::PropertyStringList",
            "Fixtures",
            "WCS",
            QT_TRANSLATE_NOOP("App::Property", "The Work Coordinate Systems for the Job"),
        )

        obj.Fixtures = ["G54"]

        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

        obj.PostProcessorOutputFile = Path.Preferences.defaultOutputFile()
        postProcessors = Path.Preferences.allEnabledPostProcessors()
        # Add empty string as a valid enumeration option
        if "" not in postProcessors:
            postProcessors = [""] + postProcessors
        obj.PostProcessor = postProcessors
        defaultPostProcessor = Path.Preferences.defaultPostProcessor()
        # Check to see if default post processor hasn't been 'lost' (This can happen when Macro dir has changed)
        if defaultPostProcessor in postProcessors:
            obj.PostProcessor = defaultPostProcessor
        else:
            obj.PostProcessor = ""
        obj.PostProcessorArgs = Path.Preferences.defaultPostProcessorArgs()
        obj.GeometryTolerance = Path.Preferences.defaultGeometryTolerance()

        self.setupOperations(obj)
        self.setupSetupSheet(obj)
        self.setupBaseModel(obj, models)
        self.setupToolTable(obj)
        self.setFromTemplateFile(obj, templateFile)
        self.setupStock(obj)

    @classmethod
    def propertyEnumerations(self, dataType="data"):
        """propertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        enums = {
            "OrderOutputBy": [
                (translate("CAM_Job", "Fixture"), "Fixture"),
                (translate("CAM_Job", "Tool"), "Tool"),
                (translate("CAM_Job", "Operation"), "Operation"),
            ],
            "JobType": [
                (translate("CAM_Job", "2D"), "2D"),
                (translate("CAM_Job", "2.5D"), "2.5D"),
                (translate("CAM_Job", "Lathe"), "Lathe"),
                (translate("CAM_Job", "Multiaxis"), "Multiaxis"),
            ],
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def setupOperations(self, obj):
        """setupOperations(obj)... setup the Operations group for the Job object."""
        # ops = FreeCAD.ActiveDocument.addObject(
        #     "Path::FeatureCompoundPython", "Operations"
        # )
        ops = FreeCAD.ActiveDocument.addObject("App::DocumentObjectGroup", "Operations")
        if ops.ViewObject:
            # ops.ViewObject.Proxy = 0
            ops.ViewObject.Visibility = True

        obj.Operations = ops
        obj.setEditorMode("Operations", 2)  # hide
        obj.setEditorMode("Placement", 2)

    def setupSetupSheet(self, obj):
        if not getattr(obj, "SetupSheet", None):
            if not hasattr(obj, "SetupSheet"):
                obj.addProperty(
                    "App::PropertyLink",
                    "SetupSheet",
                    "Base",
                    QT_TRANSLATE_NOOP(
                        "App::Property", "SetupSheet holding the settings for this job"
                    ),
                )
            obj.SetupSheet = PathSetupSheet.Create()
            if obj.SetupSheet.ViewObject:
                import Path.Base.Gui.IconViewProvider

                Path.Base.Gui.IconViewProvider.Attach(obj.SetupSheet.ViewObject, "SetupSheet")
            obj.SetupSheet.Label = "SetupSheet"
        self.setupSheet = obj.SetupSheet.Proxy

    def setupBaseModel(self, obj, models=None):
        Path.Log.track(obj.Label, models)
        addModels = False

        if not hasattr(obj, "Model"):
            obj.addProperty(
                "App::PropertyLink",
                "Model",
                "Base",
                QT_TRANSLATE_NOOP("App::Property", "The base objects for all operations"),
            )
            addModels = True
        elif obj.Model is None:
            addModels = True

        if addModels:
            model = FreeCAD.ActiveDocument.addObject("App::DocumentObjectGroup", "Model")
            if model.ViewObject:
                model.ViewObject.Visibility = False
            if models:
                model.addObjects([createModelResourceClone(obj, base) for base in models])
            obj.Model = model
            obj.Model.Label = "Model"

        if hasattr(obj, "Base"):
            Path.Log.info("Converting Job.Base to new Job.Model for {}".format(obj.Label))
            obj.Model.addObject(obj.Base)
            obj.Base = None
            obj.removeProperty("Base")

    def setupToolTable(self, obj):
        addTable = False
        if not hasattr(obj, "Tools"):
            obj.addProperty(
                "App::PropertyLink",
                "Tools",
                "Base",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Collection of all tool controllers for the job"
                ),
            )
            addTable = True
        elif obj.Tools is None:
            addTable = True

        if addTable:
            toolTable = FreeCAD.ActiveDocument.addObject("App::DocumentObjectGroup", "Tools")
            toolTable.Label = "Tools"
            if toolTable.ViewObject:
                toolTable.ViewObject.Visibility = False
            if hasattr(obj, "ToolController"):
                toolTable.addObjects(obj.ToolController)
                obj.removeProperty("ToolController")
            obj.Tools = toolTable

    def setupStock(self, obj):
        """setupStock(obj)... setup the Stock for the Job object."""
        if not obj.Stock:
            stockTemplate = Path.Preferences.defaultStockTemplate()
            if stockTemplate:
                obj.Stock = PathStock.CreateFromTemplate(obj, json.loads(stockTemplate))
            if not obj.Stock:
                obj.Stock = PathStock.CreateFromBase(obj)
        if obj.Stock.ViewObject:
            obj.Stock.ViewObject.Visibility = False

    def removeBase(self, obj, base, removeFromModel):
        if isResourceClone(obj, base, None):
            PathUtil.clearExpressionEngine(base)
            if removeFromModel:
                obj.Model.removeObject(base)
            obj.Document.removeObject(base.Name)

    def modelBoundBox(self, obj):
        return PathStock.shapeBoundBox(obj.Model.Group)

    def onDelete(self, obj, arg2=None):
        """Called by the view provider, there doesn't seem to be a callback on the obj itself."""
        Path.Log.track(obj.Label, arg2)
        doc = obj.Document

        if getattr(obj, "Operations", None):
            # the first to tear down are the ops, they depend on other resources
            Path.Log.debug("taking down ops: %s" % [o.Name for o in self.allOperations()])
            while obj.Operations.Group:
                op = obj.Operations.Group[0]
                if (
                    not op.ViewObject
                    or not hasattr(op.ViewObject.Proxy, "onDelete")
                    or op.ViewObject.Proxy.onDelete(op.ViewObject, ())
                ):
                    PathUtil.clearExpressionEngine(op)
                    doc.removeObject(op.Name)
            obj.Operations.Group = []
            doc.removeObject(obj.Operations.Name)
            obj.Operations = None

        # stock could depend on Model, so delete it first
        if getattr(obj, "Stock", None):
            Path.Log.debug("taking down stock")
            PathUtil.clearExpressionEngine(obj.Stock)
            doc.removeObject(obj.Stock.Name)
            obj.Stock = None

        # base doesn't depend on anything inside job
        if getattr(obj, "Model", None):
            for base in obj.Model.Group:
                Path.Log.debug("taking down base %s" % base.Label)
                self.removeBase(obj, base, False)
            obj.Model.Group = []
            doc.removeObject(obj.Model.Name)
            obj.Model = None

        # Tool controllers might refer to either legacy tool or toolbit
        if getattr(obj, "Tools", None):
            Path.Log.debug("taking down tool controller")
            for tc in obj.Tools.Group:
                if hasattr(tc.Tool, "BitBody") and tc.Tool.BitBody:
                    tc.Tool.BitBody.removeObjectsFromDocument()
                    doc.removeObject(tc.Tool.BitBody.Name)
                if hasattr(tc.Tool, "Proxy"):
                    PathUtil.clearExpressionEngine(tc.Tool)
                    doc.removeObject(tc.Tool.Name)
                PathUtil.clearExpressionEngine(tc)
                tc.Proxy.onDelete(tc)
                doc.removeObject(tc.Name)
            obj.Tools.Group = []
            doc.removeObject(obj.Tools.Name)
            obj.Tools = None

        # SetupSheet
        if getattr(obj, "SetupSheet", None):
            PathUtil.clearExpressionEngine(obj.SetupSheet)
            doc.removeObject(obj.SetupSheet.Name)
            obj.SetupSheet = None

        return True

    def fixupOperations(self, obj):
        if getattr(obj.Operations, "ViewObject", None):
            try:
                obj.Operations.ViewObject.DisplayMode
            except Exception:
                name = obj.Operations.Name
                label = obj.Operations.Label
                ops = FreeCAD.ActiveDocument.addObject("Path::FeatureCompoundPython", "Operations")
                ops.ViewObject.Proxy = 0
                ops.Group = obj.Operations.Group
                obj.Operations.Group = []
                obj.Operations = ops
                FreeCAD.ActiveDocument.removeObject(name)
                if label == "Unnamed":
                    ops.Label = "Operations"
                else:
                    ops.Label = label

    def onDocumentRestored(self, obj):
        self.setupBaseModel(obj)
        self.fixupOperations(obj)
        self.setupSetupSheet(obj)
        self.setupToolTable(obj)
        self.integrityCheck(obj)

        obj.setEditorMode("Operations", 2)  # hide
        obj.setEditorMode("Placement", 2)

        if hasattr(obj, "Path"):
            obj.Path = Path.Path()

        if not hasattr(obj, "CycleTime"):
            obj.addProperty(
                "App::PropertyString",
                "CycleTime",
                "Path",
                QT_TRANSLATE_NOOP("App::Property", "Operations Cycle Time Estimation"),
            )
            obj.setEditorMode("CycleTime", 1)  # read-only

        if not hasattr(obj, "Fixtures"):
            obj.addProperty(
                "App::PropertyStringList",
                "Fixtures",
                "WCS",
                QT_TRANSLATE_NOOP("App::Property", "The Work Coordinate Systems for the Job"),
            )
            obj.Fixtures = ["G54"]

        if not hasattr(obj, "OrderOutputBy"):
            obj.addProperty(
                "App::PropertyEnumeration",
                "OrderOutputBy",
                "WCS",
                QT_TRANSLATE_NOOP("App::Property", "If multiple WCS, order the output this way"),
            )
            obj.OrderOutputBy = ["Fixture", "Tool", "Operation"]

        if not hasattr(obj, "SplitOutput"):
            obj.addProperty(
                "App::PropertyBool",
                "SplitOutput",
                "Output",
                QT_TRANSLATE_NOOP("App::Property", "Split output into multiple G-code files"),
            )
            obj.SplitOutput = False

        if not hasattr(obj, "JobType"):
            obj.addProperty(
                "App::PropertyEnumeration",
                "JobType",
                "Base",
                QT_TRANSLATE_NOOP("App::Property", "Select the type of Job"),
            )
            obj.setEditorMode("JobType", 2)  # Hide

        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

    def onChanged(self, obj, prop):
        if prop == "PostProcessor" and obj.PostProcessor:
            processor = PostProcessorFactory.get_post_processor(obj, obj.PostProcessor)
            self.tooltip = processor.tooltip
            self.tooltipArgs = processor.tooltipArgs

    def baseObject(self, obj, base):
        """Return the base object, not its clone."""
        if isResourceClone(obj, base, "Model") or isResourceClone(obj, base, "Base"):
            return base.Objects[0]
        return base

    def baseObjects(self, obj):
        """Return the base objects, not their clones."""
        return [self.baseObject(obj, base) for base in obj.Model.Group]

    def resourceClone(self, obj, base):
        """resourceClone(obj, base) ... Return the resource clone for base if it exists."""
        if isResourceClone(obj, base, None):
            return base
        for b in obj.Model.Group:
            if base == b.Objects[0]:
                return b
        return None

    def setFromTemplateFile(self, obj, template):
        """setFromTemplateFile(obj, template) ... extract the properties from the given template file and assign to receiver.
        This will also create any TCs stored in the template."""
        tcs = []
        if template:
            with open(str(template), "rb") as fp:
                attrs = json.load(fp)

            if attrs.get(JobTemplate.Version) and 1 == int(attrs[JobTemplate.Version]):
                attrs = self.setupSheet.decodeTemplateAttributes(attrs)
                if attrs.get(JobTemplate.SetupSheet):
                    self.setupSheet.setFromTemplate(attrs[JobTemplate.SetupSheet])

                if attrs.get(JobTemplate.GeometryTolerance):
                    obj.GeometryTolerance = float(attrs.get(JobTemplate.GeometryTolerance))
                if attrs.get(JobTemplate.PostProcessor):
                    obj.PostProcessor = attrs.get(JobTemplate.PostProcessor)
                    if attrs.get(JobTemplate.PostProcessorArgs):
                        obj.PostProcessorArgs = attrs.get(JobTemplate.PostProcessorArgs)
                    else:
                        obj.PostProcessorArgs = ""
                if attrs.get(JobTemplate.PostProcessorOutputFile):
                    obj.PostProcessorOutputFile = attrs.get(JobTemplate.PostProcessorOutputFile)
                if attrs.get(JobTemplate.Description):
                    obj.Description = attrs.get(JobTemplate.Description)

                if attrs.get(JobTemplate.ToolController):
                    for tc in attrs.get(JobTemplate.ToolController):
                        ctrl = PathToolController.FromTemplate(tc)
                        if ctrl:
                            tcs.append(ctrl)
                        else:
                            Path.Log.debug(f"skipping TC {tc['name']}")
                if attrs.get(JobTemplate.Stock):
                    obj.Stock = PathStock.CreateFromTemplate(obj, attrs.get(JobTemplate.Stock))

                if attrs.get(JobTemplate.Fixtures):
                    obj.Fixtures = [x for y in attrs.get(JobTemplate.Fixtures) for x in y]

                if attrs.get(JobTemplate.OrderOutputBy):
                    obj.OrderOutputBy = attrs.get(JobTemplate.OrderOutputBy)

                if attrs.get(JobTemplate.SplitOutput):
                    obj.SplitOutput = attrs.get(JobTemplate.SplitOutput)

                Path.Log.debug("setting tool controllers (%d)" % len(tcs))
                if tcs:
                    obj.Tools.Group = tcs
            else:
                Path.Log.error(
                    "Unsupported PathJob template version {}".format(attrs.get(JobTemplate.Version))
                )

        if not tcs:
            self.addToolController(PathToolController.Create())

    def templateAttrs(self, obj):
        """templateAttrs(obj) ... answer a dictionary with all properties of the receiver that should be stored in a template file."""
        attrs = {}
        attrs[JobTemplate.Version] = 1
        if obj.PostProcessor:
            attrs[JobTemplate.PostProcessor] = obj.PostProcessor
            attrs[JobTemplate.PostProcessorArgs] = obj.PostProcessorArgs
            attrs[JobTemplate.Fixtures] = [{f: True} for f in obj.Fixtures]
            attrs[JobTemplate.OrderOutputBy] = obj.OrderOutputBy
            attrs[JobTemplate.SplitOutput] = obj.SplitOutput
        if obj.PostProcessorOutputFile:
            attrs[JobTemplate.PostProcessorOutputFile] = obj.PostProcessorOutputFile
        attrs[JobTemplate.GeometryTolerance] = str(obj.GeometryTolerance.Value)
        if obj.Description:
            attrs[JobTemplate.Description] = obj.Description
        return attrs

    def dumps(self):
        return None

    def loads(self, state):
        for obj in FreeCAD.ActiveDocument.Objects:
            if hasattr(obj, "Proxy") and obj.Proxy == self:
                self.obj = obj
                break
        return None

    def execute(self, obj):
        if getattr(obj, "Operations", None):
            # obj.Path = obj.Operations.Path
            self.getCycleTime()
            if hasattr(obj, "PathChanged"):
                obj.PathChanged = True

    def getCycleTime(self):
        seconds = 0

        if len(self.obj.Operations.Group):
            for op in self.obj.Operations.Group:

                # Skip inactive operations
                if PathUtil.opProperty(op, "Active") is False:
                    continue

                # Skip operations that don't have a cycletime attribute
                if PathUtil.opProperty(op, "CycleTime") is None:
                    continue

                formattedCycleTime = PathUtil.opProperty(op, "CycleTime")
                opCycleTime = 0
                try:
                    # Convert the formatted time from HH:MM:SS to just seconds
                    opCycleTime = sum(
                        x * int(t)
                        for x, t in zip([1, 60, 3600], reversed(formattedCycleTime.split(":")))
                    )
                except Exception:
                    continue

                if opCycleTime > 0:
                    seconds = seconds + opCycleTime

        cycleTimeString = time.strftime("%H:%M:%S", time.gmtime(seconds))
        self.obj.CycleTime = cycleTimeString

    def addOperation(self, op, before=None, removeBefore=False):
        group = self.obj.Operations.Group
        if op not in group:
            if before:
                try:
                    group.insert(group.index(before), op)
                    if removeBefore:
                        group.remove(before)
                except Exception as e:
                    Path.Log.error(e)
                    group.append(op)
            else:
                group.append(op)
            self.obj.Operations.Group = group
            # op.Path.Center = self.obj.Operations.Path.Center

    def nextToolNumber(self):
        # returns the next available toolnumber in the job
        group = self.obj.Tools.Group
        if len(group) > 0:
            return sorted([t.ToolNumber for t in group])[-1] + 1
        else:
            return 1

    def addToolController(self, tc):
        group = self.obj.Tools.Group
        Path.Log.debug("addToolController(%s): %s" % (tc.Label, [t.Label for t in group]))
        if tc.Name not in [str(t.Name) for t in group]:
            tc.setExpression(
                "VertRapid",
                "%s.%s"
                % (
                    self.setupSheet.expressionReference(),
                    PathSetupSheet.Template.VertRapid,
                ),
            )
            tc.setExpression(
                "HorizRapid",
                "%s.%s"
                % (
                    self.setupSheet.expressionReference(),
                    PathSetupSheet.Template.HorizRapid,
                ),
            )
            self.obj.Tools.addObject(tc)
            Notification.updateTC.emit(self.obj, tc)

    def allOperations(self):
        ops = []

        def collectBaseOps(op):
            if hasattr(op, "TypeId"):
                if op.TypeId == "Path::FeaturePython":
                    ops.append(op)
                    if hasattr(op, "Base"):
                        collectBaseOps(op.Base)
                if op.TypeId == "Path::FeatureCompoundPython":
                    ops.append(op)
                    for sub in op.Group:
                        collectBaseOps(sub)

        if getattr(self.obj, "Operations", None) and getattr(self.obj.Operations, "Group", None):
            for op in self.obj.Operations.Group:
                collectBaseOps(op)

        return ops

    def setCenterOfRotation(self, center):
        if center != self.obj.Path.Center:
            self.obj.Path.Center = center
            self.obj.Operations.Path.Center = center
            for op in self.allOperations():
                op.Path.Center = center

    def integrityCheck(self, job):
        """integrityCheck(job)... Return True if job has all expected children objects.  Attempts to restore any missing children."""
        suffix = ""
        if len(job.Name) > 3:
            suffix = job.Name[3:]

        def errorMessage(grp, job):
            Path.Log.error("{} corrupt in {} job.".format(grp, job.Name))

        if not job.Operations:
            self.setupOperations(job)
            job.Operations.Label = "Operations" + suffix
            if not job.Operations:
                errorMessage("Operations", job)
                return False
        if not job.SetupSheet:
            self.setupSetupSheet(job)
            job.SetupSheet.Label = "SetupSheet" + suffix
            if not job.SetupSheet:
                errorMessage("SetupSheet", job)
                return False
        if not job.Model:
            self.setupBaseModel(job)
            job.Model.Label = "Model" + suffix
            if not job.Model:
                errorMessage("Model", job)
                return False
        if not job.Stock:
            self.setupStock(job)
            job.Stock.Label = "Stock" + suffix
            if not job.Stock:
                errorMessage("Stock", job)
                return False
        if not job.Tools:
            self.setupToolTable(job)
            job.Tools.Label = "Tools" + suffix
            if not job.Tools:
                errorMessage("Tools", job)
                return False
        return True

    @classmethod
    def baseCandidates(cls):
        """Answer all objects in the current document which could serve as a Base for a job."""
        return sorted(
            [obj for obj in FreeCAD.ActiveDocument.Objects if cls.isBaseCandidate(obj)],
            key=lambda o: o.Label,
        )

    @classmethod
    def isBaseCandidate(cls, obj):
        """Answer true if the given object can be used as a Base for a job."""
        return PathUtil.isValidBaseObject(obj)


def Instances():
    """Instances() ... Return all Jobs in the current active document."""
    if FreeCAD.ActiveDocument:
        return [
            job
            for job in FreeCAD.ActiveDocument.Objects
            if hasattr(job, "Proxy") and isinstance(job.Proxy, ObjectJob)
        ]
    return []


def Create(name, base, templateFile=None):
    """Create(name, base, templateFile=None) ... creates a new job and all it's resources.
    If a template file is specified the new job is initialized with the values from the template."""
    if isinstance(base[0], str):
        models = []
        for baseName in base:
            models.append(FreeCAD.ActiveDocument.getObject(baseName))
    else:
        models = base
    obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.addExtension("App::GroupExtensionPython")
    obj.Proxy = ObjectJob(obj, models, templateFile)
    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import FreeCAD

from FreeCAD import Vector
from draftutils.translate import translate


WindowPresets = [
    "Fixed",
    "Open 1-pane",
    "Open 2-pane",
    "Sash 2-pane",
    "Sliding 2-pane",
    "Simple door",
    "Glass door",
    "Sliding 4-pane",
    "Awning",
    "Opening only",
]


def makeWindowPreset(windowtype, width, height, h1, h2, h3, w1, w2, o1, o2, placement=None):
    """makeWindowPreset(windowtype,width,height,h1,h2,h3,w1,w2,o1,o2,[placement]): makes a
    window object based on the given data. windowtype must be one of the names
    defined in Arch.WindowPresets"""

    if not FreeCAD.ActiveDocument:
        FreeCAD.Console.PrintError("No active document. Aborting\n")
        return

    def makeSketch(windowtype, width, height, h1, h2, h3, w1, w2, o1, o2):

        import Part
        import Sketcher

        width = float(width)
        height = float(height)
        h1 = float(h1)
        h2 = float(h2)
        h3 = float(h3)
        w1 = float(w1)
        w2 = float(w2)
        o1 = float(o1)
        o2 = float(o2)
        # width, height, h1, h2, w1, w2 cannot be null (for now)
        # TODO allow these to be null (don't create the component if so)
        if width * height * h1 * h2 * w1 * w2 == 0:
            FreeCAD.Console.PrintError(
                "Width, Height, H1, H2, W1 and W2 parameters cannot be zero. Aborting\n"
            )
            return
        # small spacing to avoid wrong auto-wires in sketch
        tol = h1 / 10
        # glass size divider
        gla = 10
        s = FreeCAD.ActiveDocument.addObject("Sketcher::SketchObject", "Sketch")

        def addRectangle(s, p1, p2, p3, p4):
            "adds a rectangle to the given sketch"

            idx = s.GeometryCount
            s.addGeometry(Part.LineSegment(p1, p2))
            s.addGeometry(Part.LineSegment(p2, p3))
            s.addGeometry(Part.LineSegment(p3, p4))
            s.addGeometry(Part.LineSegment(p4, p1))
            s.addConstraint(Sketcher.Constraint("Coincident", idx, 2, idx + 1, 1))
            s.addConstraint(Sketcher.Constraint("Coincident", idx + 1, 2, idx + 2, 1))
            s.addConstraint(Sketcher.Constraint("Coincident", idx + 2, 2, idx + 3, 1))
            s.addConstraint(Sketcher.Constraint("Coincident", idx + 3, 2, idx, 1))
            s.addConstraint(Sketcher.Constraint("Horizontal", idx))
            s.addConstraint(Sketcher.Constraint("Horizontal", idx + 2))
            s.addConstraint(Sketcher.Constraint("Vertical", idx + 1))
            s.addConstraint(Sketcher.Constraint("Vertical", idx + 3))

        def addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8):
            "adds two rectangles to the given sketch"

            addRectangle(s, p1, p2, p3, p4)
            addRectangle(s, p5, p6, p7, p8)

        def simpleFrame(s, width, height, h1, h2, tol):
            "creates a simple frame with constraints"

            p1 = Vector(h1 + tol, h1 + tol, 0)
            p2 = Vector(width - (h1 + tol), h1 + tol, 0)
            p3 = Vector(width - (h1 + tol), height - (h1 + tol), 0)
            p4 = Vector(h1 + tol, height - (h1 + tol), 0)
            p5 = Vector(h1 + h2, h1 + h2, 0)
            p6 = Vector(width - (h1 + h2), h1 + h2, 0)
            p7 = Vector(width - (h1 + h2), height - (h1 + h2), 0)
            p8 = Vector(h1 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            s.addConstraint(Sketcher.Constraint("DistanceX", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 14, 1, 10, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 14, 1, 10, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 10, 1, 6, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 10, 1, 6, 1, tol))
            if h2 == h1:
                s.renameConstraint(39, "Frame5")
                s.renameConstraint(40, "Frame6")
                s.renameConstraint(42, "Frame7")
                s.renameConstraint(41, "Frame8")

        def outerFrame(s, width, height, h1, w1, o1):

            p1 = Vector(0, 0, 0)
            p2 = Vector(width, 0, 0)
            p3 = Vector(width, height, 0)
            p4 = Vector(0, height, 0)
            p5 = Vector(h1, h1, 0)
            p6 = Vector(width - h1, h1, 0)
            p7 = Vector(width - h1, height - h1, 0)
            p8 = Vector(h1, height - h1, 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            s.addConstraint(Sketcher.Constraint("DistanceY", 1, height))  # 16
            s.addConstraint(Sketcher.Constraint("DistanceX", 0, width))  # 17
            s.renameConstraint(16, "Height")
            s.renameConstraint(17, "Width")
            s.addConstraint(Sketcher.Constraint("DistanceY", 6, 2, 2, 2, h1))
            s.addConstraint(Sketcher.Constraint("DistanceX", 2, 2, 6, 2, h1))
            s.addConstraint(Sketcher.Constraint("DistanceX", 4, 2, 0, 2, h1))
            s.addConstraint(Sketcher.Constraint("DistanceY", 0, 2, 4, 2, h1))
            s.renameConstraint(18, "Frame1")
            s.renameConstraint(19, "Frame2")
            s.renameConstraint(20, "Frame3")
            s.renameConstraint(21, "Frame4")
            s.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))
            return ["OuterFrame", "Frame", "Wire0,Wire1", str(w1 - w2) + "+V", "0.00+V"]

        def doorFrame(s, width, height, h1, w1, o1):

            p1 = Vector(0, 0, 0)
            p2 = Vector(width, 0, 0)
            p3 = Vector(width, height, 0)
            p4 = Vector(0, height, 0)
            p5 = Vector(h1, 0, 0)
            p6 = Vector(width - h1, 0, 0)
            p7 = Vector(width - h1, height - h1, 0)
            p8 = Vector(h1, height - h1, 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            s.addConstraint(Sketcher.Constraint("DistanceY", 1, height))  # 16
            s.addConstraint(Sketcher.Constraint("DistanceX", 0, width))  # 17
            s.renameConstraint(16, "Height")
            s.renameConstraint(17, "Width")
            s.addConstraint(Sketcher.Constraint("DistanceY", 6, 2, 2, 2, h1))
            s.addConstraint(Sketcher.Constraint("DistanceX", 2, 2, 6, 2, h1))
            s.addConstraint(Sketcher.Constraint("DistanceX", 4, 2, 0, 2, h1))
            s.addConstraint(Sketcher.Constraint("DistanceY", 0, 2, 4, 2, 0.0))
            s.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))
            s.renameConstraint(18, "Frame1")
            s.renameConstraint(19, "Frame2")
            s.renameConstraint(20, "Frame3")
            return ["OuterFrame", "Frame", "Wire0,Wire1", str(w1 - w2) + "+V", "0.00+V"]

        if windowtype == "Fixed":

            wp = outerFrame(s, width, height, h1, w1, o1)
            wp.extend(["Glass", "Glass panel", "Wire1", str(w1 / gla), str(w1 / 2) + "+V"])

        elif windowtype == "Open 1-pane":

            wp = outerFrame(s, width, height, h1, w1, o1)
            simpleFrame(s, width, height, h1, h2, tol)
            fw = str(w2)
            if w2 == w1:
                fw = "0.00+V"
            wp.extend(["InnerFrame", "Frame", "Wire2,Wire3,Edge8,Mode1", fw, str(o2) + "+V"])
            wp.extend(
                ["InnerGlass", "Glass panel", "Wire3", str(w2 / gla), str(o2 + w2 / 2) + "+V"]
            )

        elif windowtype == "Open 2-pane":

            wp = outerFrame(s, width, height, h1, w1, o1)
            p1 = Vector(h1 + tol, h1 + tol, 0)
            p2 = Vector((width / 2) - tol, h1 + tol, 0)
            p3 = Vector((width / 2) - tol, height - (h1 + tol), 0)
            p4 = Vector(h1 + tol, height - (h1 + tol), 0)
            p5 = Vector(h1 + h2, h1 + h2, 0)
            p6 = Vector((width / 2) - h2, h1 + h2, 0)
            p7 = Vector((width / 2) - h2, height - (h1 + h2), 0)
            p8 = Vector(h1 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            p1 = Vector((width / 2) + tol, h1 + tol, 0)
            p2 = Vector(width - (h1 + tol), h1 + tol, 0)
            p3 = Vector(width - (h1 + tol), height - (h1 + tol), 0)
            p4 = Vector((width / 2) + tol, height - (h1 + tol), 0)
            p5 = Vector((width / 2) + h2, h1 + h2, 0)
            p6 = Vector(width - (h1 + h2), h1 + h2, 0)
            p7 = Vector(width - (h1 + h2), height - (h1 + h2), 0)
            p8 = Vector((width / 2) + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 21, 2, 17, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 21, 2, 17, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 16, 1, 20, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 14, 1, 10, 1, h2))
            s.addConstraint(Sketcher.Constraint("Equal", 22, 14))
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 2, 16, 1, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceY", 10, 1, 18, 2, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceX", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 6, 1, 18, 1, -tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 6, 1, 18, 1, -tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 9, 1, 19, 2, tol))
            s.addConstraint(Sketcher.Constraint("PointOnObject", 13, 2, 22))
            s.addConstraint(Sketcher.Constraint("PointOnObject", 20, 1, 12))
            if h1 == h2:
                s.renameConstraint(55, "Frame5")
                s.renameConstraint(56, "Frame6")
                s.renameConstraint(57, "Frame7")
                s.renameConstraint(58, "Frame8")
                s.renameConstraint(59, "Frame9")
                s.renameConstraint(60, "Frame10")
            fw = str(w2)
            if w2 == w1:
                fw = "0.00+V"
            wp.extend(["LeftFrame", "Frame", "Wire2,Wire3,Edge8,Mode1", fw, str(o2) + "+V"])
            wp.extend(["LeftGlass", "Glass panel", "Wire3", str(w2 / gla), str(o2 + w2 / 2) + "+V"])
            wp.extend(["RightFrame", "Frame", "Wire4,Wire5,Edge6,Mode2", fw, str(o2) + "+V"])
            wp.extend(
                ["RightGlass", "Glass panel", "Wire5", str(w2 / gla), str(o2 + w2 / 2) + "+V"]
            )

        elif windowtype == "Sash 2-pane":

            wp = outerFrame(s, width, height, h1, w1, o1)
            p1 = Vector(h1 + tol, h1 + tol, 0)
            p2 = Vector(width - (h1 + tol), h1 + tol, 0)
            p3 = Vector(width - (h1 + tol), (height / 2) - tol, 0)
            p4 = Vector(h1 + tol, (height / 2) - tol, 0)
            p5 = Vector(h1 + h2, h1 + h2, 0)
            p6 = Vector(width - (h1 + h2), h1 + h2, 0)
            p7 = Vector(width - (h1 + h2), (height / 2) - h2, 0)
            p8 = Vector(h1 + h2, (height / 2) - h2, 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            p1 = Vector(h1 + tol, (height / 2) + tol, 0)
            p2 = Vector(width - (h1 + tol), (height / 2) + tol, 0)
            p3 = Vector(width - (h1 + tol), height - (h1 + tol), 0)
            p4 = Vector(h1 + tol, height - (h1 + tol), 0)
            p5 = Vector(h1 + h2, (height / 2) + h2, 0)
            p6 = Vector(width - (h1 + h2), (height / 2) + h2, 0)
            p7 = Vector(width - (h1 + h2), height - (h1 + h2), 0)
            p8 = Vector(h1 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 21, 2, 17, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 21, 2, 17, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 16, 2, 20, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 10, 2, 14, 2, -h2))
            s.addConstraint(Sketcher.Constraint("Equal", 23, 15))
            s.addConstraint(Sketcher.Constraint("DistanceX", 12, 1, 20, 1, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceX", 13, 2, 20, 2, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceX", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 6, 1, 18, 1, -tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 6, 1, 18, 1, -tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 10, 1, 16, 1, tol))
            s.addConstraint(Sketcher.Constraint("PointOnObject", 9, 2, 17))
            s.addConstraint(Sketcher.Constraint("PointOnObject", 16, 1, 11))
            if h1 == h2:
                s.renameConstraint(55, "Frame5")
                s.renameConstraint(56, "Frame6")
                s.renameConstraint(57, "Frame7")
                s.renameConstraint(58, "Frame8")
                s.renameConstraint(59, "Frame9")
                s.renameConstraint(60, "F10")
                s.setExpression(".Constraints.F10", "-.Constraints.Frame5")
            fw = str(w2)
            if w2 == w1:
                fw = "0.00+V"
            wp.extend(["LowerFrame", "Frame", "Wire2,Wire3", fw, str(o2 + w2) + "+V"])
            wp.extend(
                ["LowerGlass", "Glass panel", "Wire3", str(w2 / gla), str(o2 + w2 + w2 / 2) + "+V"]
            )
            wp.extend(["UpperFrame", "Frame", "Wire4,Wire5", fw, str(o2) + "+V"])
            wp.extend(
                ["UpperGlass", "Glass panel", "Wire5", str(w2 / gla), str(o2 + w2 / 2) + "+V"]
            )

        elif windowtype == "Sliding 2-pane":

            wp = outerFrame(s, width, height, h1, w1, o1)
            p1 = Vector(h1 + tol, h1 + tol, 0)
            p2 = Vector((width / 2) - tol, h1 + tol, 0)
            p3 = Vector((width / 2) - tol, height - (h1 + tol), 0)
            p4 = Vector(h1 + tol, height - (h1 + tol), 0)
            p5 = Vector(h1 + h2, h1 + h2, 0)
            p6 = Vector((width / 2) - h2, h1 + h2, 0)
            p7 = Vector((width / 2) - h2, height - (h1 + h2), 0)
            p8 = Vector(h1 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            p1 = Vector((width / 2) + tol, h1 + tol, 0)
            p2 = Vector(width - (h1 + tol), h1 + tol, 0)
            p3 = Vector(width - (h1 + tol), height - (h1 + tol), 0)
            p4 = Vector((width / 2) + tol, height - (h1 + tol), 0)
            p5 = Vector((width / 2) + h2, h1 + h2, 0)
            p6 = Vector(width - (h1 + h2), h1 + h2, 0)
            p7 = Vector(width - (h1 + h2), height - (h1 + h2), 0)
            p8 = Vector((width / 2) + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 21, 2, 17, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 21, 2, 17, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 16, 1, 20, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 14, 1, 10, 1, h2))
            s.addConstraint(Sketcher.Constraint("Equal", 22, 14))
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 2, 16, 1, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceY", 10, 1, 18, 2, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceX", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 6, 1, 18, 1, -tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 6, 1, 18, 1, -tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 9, 1, 19, 2, tol))
            s.addConstraint(Sketcher.Constraint("PointOnObject", 13, 2, 22))
            s.addConstraint(Sketcher.Constraint("PointOnObject", 12, 2, 20))
            if h1 == h2:
                s.renameConstraint(55, "Frame5")
                s.renameConstraint(56, "Frame6")
                s.renameConstraint(57, "Frame7")
                s.renameConstraint(58, "Frame8")
                s.renameConstraint(59, "Frame9")
                s.renameConstraint(60, "Frame10")
            fw = str(w2)
            if w2 == w1:
                fw = "0.00+V"
            wp.extend(["LeftFrame", "Frame", "Wire2,Wire3", fw, str(o2) + "+V"])
            wp.extend(["LeftGlass", "Glass panel", "Wire3", str(w2 / gla), str(o2 + w2 / 2) + "+V"])
            wp.extend(["RightFrame", "Frame", "Wire4,Wire5", fw, str(o2 + w2) + "+V"])
            wp.extend(
                ["RightGlass", "Glass panel", "Wire5", str(w2 / gla), str(o2 + w2 + w2 / 2) + "+V"]
            )

        elif windowtype == "Sliding 4-pane":

            wp = outerFrame(s, width, height, h1, w1, o1)
            p1 = Vector(h1 + tol, h1 + tol, 0)
            p2 = Vector(width / 4 - tol, h1 + tol, 0)
            p3 = Vector(width / 4 - tol, height - (h1 + tol), 0)
            p4 = Vector(h1 + tol, height - (h1 + tol), 0)
            p5 = Vector(h1 + h2, h1 + h2, 0)
            p6 = Vector(width / 4 - h2, h1 + h2, 0)
            p7 = Vector(width / 4 - h2, height - (h1 + h2), 0)
            p8 = Vector(h1 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            p1 = Vector(width / 4 + tol, h1 + tol, 0)
            p2 = Vector(width / 2 - tol, h1 + tol, 0)
            p3 = Vector(width / 2 - tol, height - (h1 + tol), 0)
            p4 = Vector(width / 4 + tol, height - (h1 + tol), 0)
            p5 = Vector(width / 4 + h2, h1 + h2, 0)
            p6 = Vector(width / 2 - h2, h1 + h2, 0)
            p7 = Vector(width / 2 - h2, height - (h1 + h2), 0)
            p8 = Vector(width / 4 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            p1 = Vector(width / 2 + tol, h1 + tol, 0)
            p2 = Vector(width * 3 / 4 - tol, h1 + tol, 0)
            p3 = Vector(width * 3 / 4 - tol, height - (h1 + tol), 0)
            p4 = Vector(width / 2 + tol, height - (h1 + tol), 0)
            p5 = Vector(width / 2 + h2, h1 + h2, 0)
            p6 = Vector(width * 3 / 4 - h2, h1 + h2, 0)
            p7 = Vector(width * 3 / 4 - h2, height - (h1 + h2), 0)
            p8 = Vector(width / 2 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            p1 = Vector(width * 3 / 4 + tol, h1 + tol, 0)
            p2 = Vector(width - (h1 + tol), h1 + tol, 0)
            p3 = Vector(width - (h1 + tol), height - (h1 + tol), 0)
            p4 = Vector(width * 3 / 4 + tol, height - (h1 + tol), 0)
            p5 = Vector(width * 3 / 4 + h2, h1 + h2, 0)
            p6 = Vector(width - (h1 + h2), h1 + h2, 0)
            p7 = Vector(width - (h1 + h2), height - (h1 + h2), 0)
            p8 = Vector(width * 3 / 4 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            s.addConstraint(Sketcher.Constraint("DistanceX", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 8, 2, 16, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 17, 1, 27, 2, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 24, 2, 32, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 32, 2, 4, 2, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 10, 2, 6, 2, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 17, 2, 26, 2, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceY", 25, 2, 34, 2, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 2, 16, 1, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceY", 9, 2, 18, 2, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceY", 16, 2, 24, 1, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceY", 24, 2, 32, 1, 0.0))
            s.addConstraint(Sketcher.Constraint("DistanceX", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 13, 2, 9, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 13, 2, 9, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 16, 1, 20, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 16, 1, 20, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 21, 2, 17, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 21, 2, 17, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 24, 1, 28, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 24, 1, 28, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 29, 2, 25, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 29, 2, 25, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 32, 1, 36, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 32, 1, 36, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 37, 2, 33, 2, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 37, 2, 33, 2, h2))
            s.addConstraint(Sketcher.Constraint("Equal", 14, 22))
            s.addConstraint(Sketcher.Constraint("Equal", 22, 30))
            s.addConstraint(Sketcher.Constraint("Equal", 30, 38))
            if h1 == h2:
                s.renameConstraint(100, "Frame5")
                s.renameConstraint(101, "Frame6")
                s.renameConstraint(102, "Frame7")
                s.renameConstraint(103, "Frame8")
                s.renameConstraint(104, "Frame9")
                s.renameConstraint(105, "Frame10")
                s.renameConstraint(106, "Frame11")
                s.renameConstraint(107, "Frame12")
                s.renameConstraint(108, "Frame13")
                s.renameConstraint(109, "Frame14")
                s.renameConstraint(110, "Frame15")
                s.renameConstraint(111, "Frame16")
                s.renameConstraint(112, "Frame17")
                s.renameConstraint(113, "Frame18")
                s.renameConstraint(114, "Frame19")
                s.renameConstraint(115, "Frame20")
            fw = str(w2)
            if w2 == w1:
                fw = "0.00+V"
            wp.extend(["LeftMostFrame", "Frame", "Wire2,Wire3", fw, str(o2) + "+V"])
            wp.extend(
                ["LeftMostGlass", "Glass panel", "Wire3", str(w2 / gla), str(o2 + w2 / 2) + "+V"]
            )
            wp.extend(["LeftFrame", "Frame", "Wire4,Wire5", fw, str(o2 + w2) + "+V"])
            wp.extend(
                ["LeftGlass", "Glass panel", "Wire5", str(w2 / gla), str(o2 + w2 + w2 / 2) + "+V"]
            )
            wp.extend(["RightFrame", "Frame", "Wire6,Wire7", fw, str(o2 + w2) + "+V"])
            wp.extend(
                ["RightGlass", "Glass panel", "Wire7", str(w2 / gla), str(o2 + w2 + w2 / 2) + "+V"]
            )
            wp.extend(["RightMostFrame", "Frame", "Wire8,Wire9", fw, str(o2) + "+V"])
            wp.extend(
                ["RightMostGlass", "Glass panel", "Wire9", str(w2 / gla), str(o2 + w2 / 2) + "+V"]
            )

        elif windowtype == "Awning":

            wp = outerFrame(s, width, height, h1, w1, o1)
            simpleFrame(s, width, height, h1, h2, tol)
            fw = str(w2)
            if w2 == w1:
                fw = "0.00+V"
            wp.extend(["InnerFrame", "Frame", "Wire2,Wire3,Edge7,Mode1", fw, str(o2) + "+V"])
            wp.extend(
                ["InnerGlass", "Glass panel", "Wire3", str(w2 / gla), str(o2 + w2 / 2) + "+V"]
            )

        elif windowtype == "Simple door":

            wp = doorFrame(s, width, height, h1, w1, o1)
            wp.extend(["Door", "Solid panel", "Wire1,Edge8,Mode1", str(w2), str(o2) + "+V"])

        elif windowtype == "Glass door":

            wp = doorFrame(s, width, height, h1, w1, o1)
            p1 = Vector(h1 + tol, h1 + tol, 0)
            p2 = Vector(width - (h1 + tol), h1 + tol, 0)
            p3 = Vector(width - (h1 + tol), height - (h1 + tol), 0)
            p4 = Vector(h1 + tol, height - (h1 + tol), 0)
            p5 = Vector(h1 + h2, h1 + h3, 0)
            p6 = Vector(width - (h1 + h2), h1 + h3, 0)
            p7 = Vector(width - (h1 + h2), height - (h1 + h2), 0)
            p8 = Vector(h1 + h2, height - (h1 + h2), 0)
            addFrame(s, p1, p2, p3, p4, p5, p6, p7, p8)
            s.addConstraint(Sketcher.Constraint("DistanceX", 8, 1, 12, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 8, 1, 12, 1, h3))
            s.addConstraint(Sketcher.Constraint("DistanceX", 14, 1, 10, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceY", 14, 1, 10, 1, h2))
            s.addConstraint(Sketcher.Constraint("DistanceX", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 4, 1, 8, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceX", 10, 1, 6, 1, tol))
            s.addConstraint(Sketcher.Constraint("DistanceY", 10, 1, 6, 1, tol))
            if h2 == h1:
                s.renameConstraint(39, "Frame5")
                s.renameConstraint(40, "Frame6")
                s.renameConstraint(42, "Frame7")
                s.renameConstraint(41, "Frame8")
            fw = str(w2)
            if w2 == w1:
                fw = "0.00+V"
            wp.extend(["InnerFrame", "Frame", "Wire2,Wire3,Edge8,Mode1", fw, str(o2) + "+V"])
            wp.extend(
                ["InnerGlass", "Glass panel", "Wire3", str(w2 / gla), str(o2 + w2 / 2) + "+V"]
            )

        elif windowtype == "Opening only":

            wp = []
            p1 = Vector(0, 0, 0)
            p2 = Vector(width, 0, 0)
            p3 = Vector(width, height, 0)
            p4 = Vector(0, height, 0)
            addRectangle(s, p1, p2, p3, p4)
            s.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))
            s.addConstraint(Sketcher.Constraint("DistanceX", 0, width))
            s.addConstraint(Sketcher.Constraint("DistanceY", 1, height))
            s.renameConstraint(9, "Width")
            s.renameConstraint(10, "Height")

        return (s, wp)

    if windowtype in WindowPresets:
        import Arch

        default = makeSketch(windowtype, width, height, h1, h2, h3, w1, w2, o1, o2)
        FreeCAD.ActiveDocument.recompute()
        if default:
            if placement:
                default[0].Placement = placement
                FreeCAD.ActiveDocument.recompute()
            obj = Arch.makeWindow(default[0], width, height, default[1])
            obj.Preset = WindowPresets.index(windowtype) + 1
            obj.Frame = w2
            obj.Offset = o1
            obj.Placement = FreeCAD.Placement()  # unable to find where this bug comes from...
            if "door" in windowtype.lower():
                obj.IfcType = "Door"
                obj.Label = translate("Arch", "Door")
            elif "opening" in windowtype.lower():
                obj.IfcType = "Opening Element"
                obj.Label = translate("Arch", "Opening")
            FreeCAD.ActiveDocument.recompute()
            return obj

    print("Arch: Unknown window type")



# SPDX-License-Identifier: LGPL-2.1-or-later

# *****************************************************************************
# *                                                                           *
# *   Copyright (c) 2014 Jonathan Wiedemann <wood.galaxy@gmail.com> (cutplan) *
# *   Copyright (c) 2019 Jerome Laverroux <jerome.laverroux@free.fr> (cutline)*
# *   Copyright (c) 2023 FreeCAD Project Association                          *
# *                                                                           *
# *   This file is part of FreeCAD.                                           *
# *                                                                           *
# *   FreeCAD is free software: you can redistribute it and/or modify it      *
# *   under the terms of the GNU Lesser General Public License as             *
# *   published by the Free Software Foundation, either version 2.1 of the    *
# *   License, or (at your option) any later version.                         *
# *                                                                           *
# *   FreeCAD is distributed in the hope that it will be useful, but          *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of              *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
# *   Lesser General Public License for more details.                         *
# *                                                                           *
# *   You should have received a copy of the GNU Lesser General Public        *
# *   License along with FreeCAD. If not, see                                 *
# *   <https://www.gnu.org/licenses/>.                                        *
# *                                                                           *
# *****************************************************************************

__title__ = "FreeCAD CutPlane"
__author__ = "Jonathan Wiedemann"
__url__ = "https://www.freecad.org"

## @package ArchCutPlane
#  \ingroup ARCH
#  \brief The Cut plane object and tools
#
#  This module handles the Cut Plane object

import FreeCAD
import ArchCommands
import Draft
import Part

if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    # \endcond


# _getShapes(FreeCADGui.Selection.getSelectionEx("", 0))
def _getShapes(sels):
    """Check and process the user selection.
    Returns a tuple: (baseObj, baseShp, cutterShp).
    baseShp and cutterShp are in the global coordinate system, cutterShp is a planar face.
    If the selection is not valid one or more items in the tuple will be `None`.
    """
    if not sels:
        return None, None, None
    objs = []
    needSubEle = False
    for sel in sels:
        for sub in sel.SubElementNames if sel.SubElementNames else [""]:
            objs.append(Part.getShape(sel.Object, sub, needSubElement=needSubEle, retType=1))
            needSubEle = True
    if len(objs) != 2:
        return None, None, None
    baseShp, _, baseObj = objs[0]
    cutterShp, _, _ = objs[1]
    if baseShp.isNull():
        return baseObj, None, None
    if cutterShp.isNull():
        return baseObj, baseShp, None
    if cutterShp.ShapeType == "Edge":
        if isinstance(cutterShp.Curve, Part.Line):
            cutterShp = _extrudeEdge(cutterShp)
        else:
            try:
                cutterShp = Part.Face(Part.Wire(cutterShp))
            except Part.OCCError:
                pass
    elif cutterShp.ShapeType == "Wire":
        if len(cutterShp.Edges) == 1 and isinstance(cutterShp.Edges[0].Curve, Part.Line):
            cutterShp = _extrudeEdge(cutterShp.Edges[0])
        else:
            try:
                cutterShp = Part.Face(cutterShp)
            except Part.OCCError:
                pass
    if not cutterShp.Faces and cutterShp.Vertexes:
        plane = cutterShp.findPlane()
        if plane is not None:
            # Directly creating a face from the plane results in an almost
            # endless face that ArchCommands.getCutVolume() cannot handle.
            # We therefore create a small triangular face.
            pt_main = cutterShp.Vertexes[0].Point
            mtx = plane.Rotation.toMatrix()
            pt_u = mtx.col(0) + pt_main
            pt_v = mtx.col(1) + pt_main
            cutterShp = Part.Face(Part.makePolygon([pt_main, pt_u, pt_v, pt_main]))
    # _extrudeEdge can create a face with a zero area (if the edge is parallel to the WP normal):
    if not cutterShp.Faces or cutterShp.Faces[0].Area < 1e-6 or cutterShp.findPlane() is None:
        return baseObj, baseShp, None
    return baseObj, baseShp, cutterShp.Faces[0]


def _extrudeEdge(edge):
    """Exrude an edge along the WP normal"""
    import WorkingPlane

    return edge.extrude(WorkingPlane.get_working_plane().axis)


def cutComponentwithPlane(baseObj, cutterShp=None, side=0):
    """cut an object with a plane defined by a face.

    Parameters
    ----------
    baseObj: Part::FeaturePython object or selection set (a list of Gui::SelectionObject objects)
        Object to be cut or a selection set: `FreeCADGui.Selection.getSelectionEx("", 0)`.
        If a selection set is provided it should contain baseObj and cutterShp, in that order.

    cutterShp: Part.Shape, optional
        Defaults to `None` in which case cutterShp should be in the baseObj selection set.
        Either a face or an edge. An edge is extruded along the Draft working plane normal.
        The shape should be in the global coordinate system.

    side: 0 or 1, optional
        Defaults to 0.
        Behind = 0, front = 1.
    """
    if (
        isinstance(baseObj, list)
        and len(baseObj) >= 1
        and baseObj[0].isDerivedFrom("Gui::SelectionObject")
    ):
        baseObj, baseShp, cutterShp = _getShapes(baseObj)
        baseParent = baseObj.getParentGeoFeatureGroup()
    else:
        baseShp = baseObj.Shape
        baseParent = baseObj.getParentGeoFeatureGroup()
        if baseParent is not None:
            baseShp = baseShp.transformGeometry(baseParent.getGlobalPlacement().toMatrix())

    if cutterShp.ShapeType != "Face":
        cutterShp = _extrudeEdge(cutterShp)

    cutVolume = ArchCommands.getCutVolume(cutterShp, baseShp)
    cutVolume = cutVolume[2] if side == 0 else cutVolume[1]
    if cutVolume:
        obj = FreeCAD.ActiveDocument.addObject("Part::Feature", "CutVolume")
        if baseParent is not None:
            cutVolume.Placement = baseParent.getGlobalPlacement().inverse()
        obj.Shape = Part.Compound([cutVolume])
        if baseParent is not None:
            baseParent.addObject(obj)
        if "Additions" in baseObj.PropertiesList:
            ArchCommands.removeComponents(obj, baseObj)  # Also changes the obj colors.
        else:
            Draft.format_object(obj, baseObj)
            cutObj = FreeCAD.ActiveDocument.addObject("Part::Cut", "CutPlane")
            if baseParent is not None:
                baseParent.addObject(cutObj)
            cutObj.Base = baseObj
            cutObj.Tool = obj



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM material mechanical nonlinear document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package material_mechanicalnonlinear
#  \ingroup FEM
#  \brief nonlinear mechanical material object

from . import base_fempythonobject


class MaterialMechanicalNonlinear(base_fempythonobject.BaseFemPythonObject):
    """
    The MaterialMechanicalNonlinear object
    """

    Type = "Fem::MaterialMechanicalNonlinear"

    def __init__(self, obj):
        super().__init__(obj)
        self.add_properties(obj)

        obj.addExtension("App::SuppressibleExtensionPython")

    def onDocumentRestored(self, obj):

        # YieldPoints was (until 0.19) stored as 3 separate variables. Consolidate them if present.
        yield_points = []
        if hasattr(obj, "YieldPoint1"):
            if obj.YieldPoint1:
                yield_points.append(obj.YieldPoint1)
            obj.removeProperty("YieldPoint1")
            if hasattr(obj, "YieldPoint2"):
                if obj.YieldPoint2:
                    yield_points.append(obj.YieldPoint2)
                obj.removeProperty("YieldPoint2")
            if hasattr(obj, "YieldPoint3"):
                if obj.YieldPoint3:
                    yield_points.append(obj.YieldPoint3)
                obj.removeProperty("YieldPoint3")

        self.add_properties(obj)
        if yield_points:
            obj.YieldPoints = yield_points

        if not obj.hasExtension("App::SuppressibleExtensionPython"):
            obj.addExtension("App::SuppressibleExtensionPython")
        # TODO: If in the future more nonlinear options are added, update choices here.

    def add_properties(self, obj):

        # this method is called from onDocumentRestored
        # thus only add and or set a attribute
        # if the attribute does not exist

        if not hasattr(obj, "LinearBaseMaterial"):
            obj.addProperty(
                "App::PropertyLink",
                "LinearBaseMaterial",
                "Base",
                "Set the linear material the nonlinear builds upon.",
            )
            obj.setPropertyStatus("LinearBaseMaterial", "LockDynamic")

        if not hasattr(obj, "MaterialModelNonlinearity"):
            choices_nonlinear_material_models = ["isotropic hardening", "kinematic hardening"]
            obj.addProperty(
                "App::PropertyEnumeration",
                "MaterialModelNonlinearity",
                "Fem",
                "Set the type on nonlinear material model",
            )
            obj.setPropertyStatus("MaterialModelNonlinearity", "LockDynamic")
            obj.MaterialModelNonlinearity = choices_nonlinear_material_models
            obj.MaterialModelNonlinearity = choices_nonlinear_material_models[0]

        if (
            hasattr(obj, "MaterialModelNonlinearity")
            and obj.MaterialModelNonlinearity == "simple hardening"
        ):
            updated_choices_nonlinear_material_models = [
                "isotropic hardening",
                "kinematic hardening",
            ]
            obj.MaterialModelNonlinearity = updated_choices_nonlinear_material_models
            obj.MaterialModelNonlinearity = updated_choices_nonlinear_material_models[0]

        if not hasattr(obj, "YieldPoints"):
            obj.addProperty(
                "App::PropertyStringList",
                "YieldPoints",
                "Fem",
                "Set stress and strain for yield points as a list of strings, "
                'each point "stress, plastic strain"',
            )
            obj.setPropertyStatus("YieldPoints", "LockDynamic")
            obj.YieldPoints = []



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Arch Curtain Wall"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchCurtainWall
#  \ingroup ARCH
#  \brief The Curtain Wall object and tools
#
#  This module provides tools to build Curtain wall objects.

"""
Curtain wall tool

Abstract: Curtain walls need a surface to work on (base).
They then divide each face of that surface into quads,
using the face parameters grid.

The vertical lines can then receive one type of profile
(vertical mullions), the horizontal ones another
(horizontal mullions), and the quads a third (panels).

We then have two cases, depending on each quad: Either the
four corners of each quad are coplanar, in which case the
panel filling is rectangular, or they don't, in which case
the facet is triangulated and receives a third mullion
(diagonal mullion).
"""

import math

import FreeCAD
import ArchCommands
import ArchComponent
import DraftVecUtils

from draftutils import params

if FreeCAD.GuiUp:
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond

ANGLETOLERANCE = 0.67  # vectors with angles below this are considered going in same dir


class CurtainWall(ArchComponent.Component):
    "The curtain wall object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "CurtainWall"
        self.setProperties(obj)
        obj.IfcType = "Curtain Wall"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        vsize = 50
        hsize = 50
        if not "Host" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Host",
                "CurtainWall",
                QT_TRANSLATE_NOOP("App::Property", "An optional host object for this curtain wall"),
                locked=True,
            )
        if not "Height" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Height",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The height of the curtain wall, if based on an edge"
                ),
                locked=True,
            )
            obj.Height = params.get_param_arch("WallHeight")
        if not "VerticalMullionNumber" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "VerticalMullionNumber",
                "CurtainWall",
                QT_TRANSLATE_NOOP("App::Property", "The number of vertical mullions"),
                locked=True,
            )
            obj.setEditorMode("VerticalMullionNumber", 1)
        if not "VerticalMullionAlignment" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "VerticalMullionAlignment",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If the profile of the vertical mullions get aligned with the surface or not",
                ),
                locked=True,
            )
        if not "VerticalSections" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "VerticalSections",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The number of vertical sections of this curtain wall"
                ),
                locked=True,
            )
            obj.VerticalSections = 1
        if "VerticalMullionSize" in pl:
            # obsolete
            vsize = obj.VerticalMullionSize.Value
            obj.removeProperty("VerticalMullionSize")
        if not "VerticalMullionHeight" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "VerticalMullionHeight",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The height of the vertical mullions profile, if no profile is used",
                ),
                locked=True,
            )
            obj.VerticalMullionHeight = vsize
        if not "VerticalMullionWidth" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "VerticalMullionWidth",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The width of the vertical mullions profile, if no profile is used",
                ),
                locked=True,
            )
            obj.VerticalMullionWidth = vsize
        if not "VerticalMullionProfile" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "VerticalMullionProfile",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "A profile for vertical mullions (disables vertical mullion size)",
                ),
                locked=True,
            )
        if not "HorizontalMullionNumber" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "HorizontalMullionNumber",
                "CurtainWall",
                QT_TRANSLATE_NOOP("App::Property", "The number of horizontal mullions"),
                locked=True,
            )
            obj.setEditorMode("HorizontalMullionNumber", 1)
        if not "HorizontalMullionAlignment" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "HorizontalMullionAlignment",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If the profile of the horizontal mullions gets aligned with the surface or not",
                ),
                locked=True,
            )
        if not "HorizontalSections" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "HorizontalSections",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The number of horizontal sections of this curtain wall"
                ),
                locked=True,
            )
            obj.HorizontalSections = 1
        if "HorizontalMullionSize" in pl:
            # obsolete
            hsize = obj.HorizontalMullionSize.Value
            obj.removeProperty("HorizontalMullionSize")
        if not "HorizontalMullionHeight" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "HorizontalMullionHeight",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The height of the horizontal mullions profile, if no profile is used",
                ),
                locked=True,
            )
            obj.HorizontalMullionHeight = hsize
        if not "HorizontalMullionWidth" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "HorizontalMullionWidth",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The width of the horizontal mullions profile, if no profile is used",
                ),
                locked=True,
            )
            obj.HorizontalMullionWidth = hsize
        if not "HorizontalMullionProfile" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "HorizontalMullionProfile",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "A profile for horizontal mullions (disables horizontal mullion size)",
                ),
                locked=True,
            )
        if not "DiagonalMullionNumber" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "DiagonalMullionNumber",
                "CurtainWall",
                QT_TRANSLATE_NOOP("App::Property", "The number of diagonal mullions"),
                locked=True,
            )
            obj.setEditorMode("DiagonalMullionNumber", 1)
        if not "DiagonalMullionSize" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "DiagonalMullionSize",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The size of the diagonal mullions, if any, if no profile is used",
                ),
                locked=True,
            )
            obj.DiagonalMullionSize = 50
        if not "DiagonalMullionProfile" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "DiagonalMullionProfile",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "A profile for diagonal mullions, if any (disables horizontal mullion size)",
                ),
                locked=True,
            )
        if not "PanelNumber" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "PanelNumber",
                "CurtainWall",
                QT_TRANSLATE_NOOP("App::Property", "The number of panels"),
                locked=True,
            )
            obj.setEditorMode("PanelNumber", 1)
        if not "PanelThickness" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "PanelThickness",
                "CurtainWall",
                QT_TRANSLATE_NOOP("App::Property", "The thickness of the panels"),
                locked=True,
            )
            obj.PanelThickness = 20
        if not "SwapHorizontalVertical" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "SwapHorizontalVertical",
                "CurtainWall",
                QT_TRANSLATE_NOOP("App::Property", "Swaps horizontal and vertical lines"),
                locked=True,
            )
        if not "Refine" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "Refine",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Perform subtractions between components so none overlap"
                ),
                locked=True,
            )
        if not "CenterProfiles" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "CenterProfiles",
                "CurtainWall",
                QT_TRANSLATE_NOOP("App::Property", "Centers the profile over the edges or not"),
                locked=True,
            )
            obj.CenterProfiles = True
        if not "VerticalDirection" in pl:
            obj.addProperty(
                "App::PropertyVector",
                "VerticalDirection",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The vertical direction reference to be used by this object to deduce vertical/horizontal directions. Keep it close to the actual vertical direction of your curtain wall",
                ),
                locked=True,
            )
            obj.VerticalDirection = FreeCAD.Vector(0, 0, 1)
        if not "OverrideEdges" in pl:  # PropertyStringList
            obj.addProperty(
                "App::PropertyStringList",
                "OverrideEdges",
                "CurtainWall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Input are index numbers of edges of Base ArchSketch/Sketch geometries (in Edit mode).  Selected edges are used to create the shape of this Arch Curtain Wall (instead of using all edges by default).  [ENHANCED by ArchSketch] GUI 'Edit Curtain Wall' Tool is provided in external Add-on ('SketchArch') to let users to select the edges interactively.  'Toponaming-Tolerant' if ArchSketch is used in Base (and SketchArch Add-on is installed).  Warning : Not 'Toponaming-Tolerant' if just Sketch is used. Property is ignored if Base ArchSketch provided the selected edges.",
                ),
                locked=True,
            )

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

    def loads(self, state):

        self.Type = "CurtainWall"

    def onChanged(self, obj, prop):

        ArchComponent.Component.onChanged(self, obj, prop)

    def execute(self, obj):

        if self.clone(obj):
            return
        if not self.ensureBase(obj):
            return

        import Part
        import DraftGeomUtils

        pl = obj.Placement

        # test properties
        if not obj.Base:
            FreeCAD.Console.PrintLog(obj.Label + ": no base\n")
            return
        if not hasattr(obj.Base, "Shape"):
            FreeCAD.Console.PrintLog(obj.Label + ": invalid base\n")
            return
        if obj.VerticalMullionProfile:
            if not hasattr(obj.VerticalMullionProfile, "Shape"):
                FreeCAD.Console.PrintLog(obj.Label + ": invalid vertical mullion profile\n")
                return
        if obj.HorizontalMullionProfile:
            if not hasattr(obj.HorizontalMullionProfile, "Shape"):
                FreeCAD.Console.PrintLog(obj.Label + ": invalid horizontal mullion profile\n")
                return
        if obj.DiagonalMullionProfile:
            if not hasattr(obj.DiagonalMullionProfile, "Shape"):
                FreeCAD.Console.PrintLog(obj.Label + ": invalid diagonal mullion profile\n")
                return

        facets = []
        faces = []

        curtainWallBaseShapeEdges = None
        curtainWallEdges = None
        if obj.Base.Shape.Faces:
            faces = obj.Base.Shape.Faces
        elif obj.Height.Value and obj.VerticalDirection.Length:
            ext = FreeCAD.Vector(obj.VerticalDirection)
            ext.normalize()
            ext = ext.multiply(obj.Height.Value)
            if hasattr(obj.Base, "Proxy"):
                if hasattr(obj.Base.Proxy, "getCurtainWallBaseShapeEdgesInfo"):
                    curtainWallBaseShapeEdges = obj.Base.Proxy.getCurtainWallBaseShapeEdgesInfo(
                        obj.Base
                    )
                    # returned a {dict}
            # get curtain wall edges (not wires); use original edges if getCurtainWallBaseShapeEdges() provided none
            if (
                curtainWallBaseShapeEdges
            ):  # would be false (none) if SketchArch Add-on is not installed, or base ArchSketch does not have the edges stored / input by user
                curtainWallEdges = curtainWallBaseShapeEdges.get("curtainWallEdges")
            elif obj.Base.isDerivedFrom("Sketcher::SketchObject"):
                skGeom = obj.Base.GeometryFacadeList
                skGeomEdges = []
                skPlacement = obj.Base.Placement  # Get Sketch's placement to restore later
                for ig, geom in enumerate(skGeom):
                    if (not obj.OverrideEdges and not geom.Construction) or str(
                        ig
                    ) in obj.OverrideEdges:
                        # support Line, Arc, Circle, Ellipse for Sketch
                        # as Base at the moment
                        if isinstance(
                            geom.Geometry, (Part.LineSegment, Part.Circle, Part.ArcOfCircle)
                        ):
                            skGeomEdgesI = geom.Geometry.toShape()
                            skGeomEdges.append(skGeomEdgesI)
                curtainWallEdges = []
                for edge in skGeomEdges:
                    edge.Placement = edge.Placement.multiply(skPlacement)
                    curtainWallEdges.append(edge)
            if not curtainWallEdges:
                curtainWallEdges = obj.Base.Shape.Edges
            if curtainWallEdges:
                faces = [edge.extrude(ext) for edge in curtainWallEdges]
        if not faces:
            FreeCAD.Console.PrintLog(obj.Label + ": unable to build base faces\n")
            return

        # subdivide the faces into quads
        for face in faces:

            fp = face.ParameterRange

            # guessing horizontal/vertical directions
            vdir = obj.VerticalDirection
            if not vdir.Length:
                vdir = FreeCAD.Vector(0, 0, 1)
            vdir.normalize()

            # Check if face if vertical in the first place
            # Fix issue in 'Curtain wall vertical/horizontal mullion mix-up'
            # https://github.com/FreeCAD/FreeCAD/issues/21845
            #
            face_plane = face.findPlane()

            if face_plane:
                if (
                    -0.001 < face_plane.Axis[2] < 0.001
                ):  # i.e. face is vertical (normal pointing horizon)
                    faceVert = True
                    # Support 'Swap Horizontal Vertical'
                    # See issue 'Swap Horizontal Vertical does not work'
                    # https://github.com/FreeCAD/FreeCAD/issues/21866
                    if obj.SwapHorizontalVertical:
                        vertsec = obj.HorizontalSections
                        horizsec = obj.VerticalSections
                    else:
                        vertsec = obj.VerticalSections
                        horizsec = obj.HorizontalSections
                else:
                    faceVert = False
            else:
                # Curve extruded face (surface) seems return no Plane - need more robust algorithm
                faceVert = False

            # Guess algorithm if face is not vertical
            if not faceVert:
                # TODO 2025.6.15 : Need a more robust algorithm below
                # See issue 'Curtain wall vertical/horizontal mullion mix-up'
                # https://github.com/FreeCAD/FreeCAD/issues/21845
                # Partially improved by checking 'if face is vertical' above
                #
                basevector = face.valueAt(fp[1], fp[3]).sub(face.valueAt(fp[0], fp[2]))
                bv_angle = basevector.getAngle(vdir)
                if (bv_angle <= math.pi / 2 + ANGLETOLERANCE) and (
                    bv_angle >= math.pi / 2 - ANGLETOLERANCE
                ):
                    facedir = True
                    if obj.SwapHorizontalVertical:
                        vertsec = obj.HorizontalSections
                        horizsec = obj.VerticalSections
                    else:
                        vertsec = obj.VerticalSections
                        horizsec = obj.HorizontalSections
                else:
                    facedir = False
                    if obj.SwapHorizontalVertical:
                        vertsec = obj.VerticalSections
                        horizsec = obj.HorizontalSections
                    else:
                        vertsec = obj.HorizontalSections
                        horizsec = obj.VerticalSections

            hstep = fp[1] - fp[0]
            if vertsec:
                hstep = hstep / vertsec
            vstep = fp[3] - fp[2]
            if horizsec:
                vstep = vstep / horizsec

            # construct facets
            for i in range(vertsec or 1):
                for j in range(horizsec or 1):
                    p0 = face.valueAt(fp[0] + i * hstep, fp[2] + j * vstep)
                    p1 = face.valueAt(fp[0] + (i + 1) * hstep, fp[2] + j * vstep)
                    p2 = face.valueAt(fp[0] + (i + 1) * hstep, fp[2] + (j + 1) * vstep)
                    p3 = face.valueAt(fp[0] + i * hstep, fp[2] + (j + 1) * vstep)
                    facet = Part.Face(Part.makePolygon([p0, p1, p2, p3, p0]))
                    facets.append(facet)

        if not facets:
            FreeCAD.Console.PrintLog(obj.Label + ": failed to subdivide shape\n")
            return

        baseshape = Part.makeShell(facets)

        # make edge/normal relation table
        edgetable = {}
        for face in baseshape.Faces:
            for edge in face.Edges:
                ec = edge.hashCode()
                if ec in edgetable:
                    edgetable[ec].append(face)
                else:
                    edgetable[ec] = [face]
        self.edgenormals = {}
        for ec, faces in edgetable.items():
            if len(faces) == 1:
                self.edgenormals[ec] = faces[0].normalAt(0, 0)
            else:
                n = faces[0].normalAt(0, 0).add(faces[1].normalAt(0, 0))
                if n.Length > 0.001:
                    n.normalize()
                else:
                    # adjacent faces have same normals
                    n = faces[0].normalAt(0, 0)
                self.edgenormals[ec] = n

        # sort edges between vertical/horizontal
        hedges = []
        vedges = []
        for edge in baseshape.Edges:
            v = edge.Vertexes[-1].Point.sub(edge.Vertexes[0].Point)
            a = v.getAngle(vdir)
            if (a <= math.pi / 2 + ANGLETOLERANCE) and (a >= math.pi / 2 - ANGLETOLERANCE):
                hedges.append(edge)
            else:
                vedges.append(edge)

        # construct vertical mullions
        vmullions = []
        vprofile = self.getMullionProfile(obj, "Vertical")
        if vprofile and vertsec:
            for vedge in vedges:
                vn = self.edgenormals[vedge.hashCode()]
                if (vn.x != 0) or (vn.y != 0):
                    avn = FreeCAD.Vector(vn.x, vn.y, 0)
                    rot = FreeCAD.Rotation(FreeCAD.Vector(0, -1, 0), avn)
                else:
                    rot = FreeCAD.Rotation()
                if obj.VerticalMullionAlignment:
                    ev = vedge.Vertexes[-1].Point.sub(vedge.Vertexes[0].Point)
                    rot = FreeCAD.Rotation(FreeCAD.Vector(1, 0, 0), ev).multiply(rot)
                vmullions.append(self.makeMullion(vedge, vprofile, rot, obj.CenterProfiles))

        # construct horizontal mullions
        hmullions = []
        hprofile = self.getMullionProfile(obj, "Horizontal")
        if hprofile and horizsec:
            for hedge in hedges:
                rot = FreeCAD.Rotation(FreeCAD.Vector(0, 1, 0), -90)
                vn = self.edgenormals[hedge.hashCode()]
                if (vn.x != 0) or (vn.y != 0):
                    avn = FreeCAD.Vector(vn.x, vn.y, 0)
                    rot = FreeCAD.Rotation(FreeCAD.Vector(0, -1, 0), avn).multiply(rot)
                    if obj.HorizontalMullionAlignment:
                        rot = FreeCAD.Rotation(avn, vn).multiply(rot)
                hmullions.append(self.makeMullion(hedge, hprofile, rot, obj.CenterProfiles))

        # construct panels
        panels = []
        dedges = []
        if obj.PanelThickness.Value:
            for face in baseshape.Faces:
                verts = [v.Point for v in face.OuterWire.OrderedVertexes]
                if DraftGeomUtils.isPlanar(verts):
                    panel = self.makePanel(verts, obj.PanelThickness.Value)
                    panels.append(panel)
                elif len(verts) == 4:
                    verts1 = [verts[0], verts[1], verts[2]]
                    panel = self.makePanel(verts1, obj.PanelThickness.Value)
                    panels.append(panel)
                    verts2 = [verts[0], verts[2], verts[3]]
                    panel = self.makePanel(verts2, obj.PanelThickness.Value)
                    panels.append(panel)
                    dedges.append(Part.makeLine(verts[0], verts[2]))

        # construct diagonal mullions
        dmullions = []
        if dedges:
            n = dedges[0].Vertexes[-1].Point.sub(dedges[0].Point)
            dprofile = self.getMullionProfile(obj, "Diagonal")
            if dprofile:
                for dedge in dedges:
                    rot = FreeCAD.Rotation(
                        FreeCAD.Vector(0, 0, 1),
                        dedge.Vertexes[-1].Point.sub(dedge.Vertexes[0].Point),
                    )
                    dmullions.append(self.makeMullion(dedge, dprofile, rot, obj.CenterProfiles))

        # perform subtractions
        if obj.Refine:
            subvmullion = None
            subhmullion = None
            subdmullion = None
            if vmullions:
                subvmullion = vmullions[0].copy()
                for m in vmullions[1:]:
                    subvmullion = subvmullion.fuse(m)
            if hmullions:
                subhmullion = hmullions[0].copy()
                for m in hmullions[1:]:
                    subhmullion = subhmullion.fuse(m)
            if dmullions:
                subdmullion = dmullions[0].copy()
                for m in dmullions[1:]:
                    subdmullion = subdmullion.fuse(m)
            if subvmullion:
                hmullions = [m.cut(subvmullion) for m in hmullions]
                if subhmullion:
                    dmullions = [m.cut(subvmullion) for m in dmullions]
                    dmullions = [m.cut(subhmullion) for m in dmullions]
                    panels = [m.cut(subvmullion) for m in panels]
                    panels = [m.cut(subhmullion) for m in panels]
                    if subdmullion:
                        panels = [m.cut(subdmullion) for m in panels]

        # mount shape
        obj.VerticalMullionNumber = len(vmullions)
        obj.HorizontalMullionNumber = len(hmullions)
        obj.DiagonalMullionNumber = len(dmullions)
        obj.PanelNumber = len(panels)
        shape = Part.makeCompound(vmullions + hmullions + dmullions + panels)
        shape = self.processSubShapes(obj, shape, pl)
        self.applyShape(obj, shape, pl)

    def makePanel(self, verts, thickness):
        """creates a panel from face points and thickness"""

        import Part

        panel = Part.Face(Part.makePolygon(verts + [verts[0]]))
        n = panel.normalAt(0, 0)
        n.multiply(thickness)
        panel = panel.extrude(n)
        return panel

    def makeMullion(self, edge, profile, rotation, recenter=False):
        """creates a mullions from an edge and a profile"""

        center = FreeCAD.Vector(0, 0, 0)
        if recenter:
            if hasattr(profile, "CenterOfMass"):
                center = profile.CenterOfMass
            elif hasattr(profile, "BoundBox"):
                center = profile.BoundBox.Center
        p0 = edge.Vertexes[0].Point
        p1 = edge.Vertexes[-1].Point
        mullion = profile.copy()
        if rotation:
            mullion = mullion.rotate(center, rotation.Axis, math.degrees(rotation.Angle))
        mullion = mullion.translate(p0.sub(center))
        mullion = mullion.extrude(p1.sub(p0))
        return mullion

    def getMullionProfile(self, obj, direction):
        """returns a profile shape already properly oriented, ready for extrude"""

        import Part
        import DraftGeomUtils

        prof = getattr(obj, direction + "MullionProfile")
        proh = getattr(obj, direction + "MullionHeight").Value
        prow = getattr(obj, direction + "MullionWidth").Value
        if prof:
            profile = prof.Shape.copy()
        else:
            if (not proh) or (not prow):
                return None
            profile = Part.Face(Part.makePlane(prow, proh, FreeCAD.Vector(-prow / 2, -proh / 2, 0)))
        return profile

    def getProjectedLength(self, v, ref):
        """gets a signed length from projecting a vector on another"""

        proj = DraftVecUtils.project(v, ref)
        if proj.getAngle(ref) < 1:
            return proj.Length
        else:
            return -proj.Length


class ViewProviderCurtainWall(ArchComponent.ViewProviderComponent):
    "A View Provider for the CurtainWall object"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_CurtainWall_Tree.svg"

    def updateData(self, obj, prop):

        if prop == "Shape":
            self.colorize(obj, force=True)

    def onChanged(self, vobj, prop):

        if (prop in ["DiffuseColor", "Transparency"]) and vobj.Object:
            self.colorize(vobj.Object)
        elif prop == "ShapeColor":
            self.colorize(vobj.Object, force=True)
        ArchComponent.ViewProviderComponent.onChanged(self, vobj, prop)

    def colorize(self, obj, force=False):
        "setting different part colors"

        if not obj.Shape or not obj.Shape.Solids:
            return
        if not obj.ViewObject:
            return
        basecolor = obj.ViewObject.ShapeColor
        basetransparency = obj.ViewObject.Transparency / 100.0
        panelcolor = ArchCommands.getDefaultColor("WindowGlass")
        paneltransparency = 0.7
        if hasattr(obj, "Material") and obj.Material and hasattr(obj.Material, "Materials"):
            if obj.Material.Names:
                if "Frame" in obj.Material.Names:
                    mat = obj.Material.Materials[obj.Material.Names.index("Frame")]
                    if ("DiffuseColor" in mat.Material) and ("(" in mat.Material["DiffuseColor"]):
                        basecolor = tuple(
                            [float(f) for f in mat.Material["DiffuseColor"].strip("()").split(",")]
                        )
                if "Glass panel" in obj.Material.Names:
                    mat = obj.Material.Materials[obj.Material.Names.index("Glass panel")]
                    if ("DiffuseColor" in mat.Material) and ("(" in mat.Material["DiffuseColor"]):
                        panelcolor = tuple(
                            [float(f) for f in mat.Material["DiffuseColor"].strip("()").split(",")]
                        )
                    if "Transparency" in mat.Material:
                        paneltransparency = float(mat.Material["Transparency"]) / 100.0
                elif "Solid panel" in obj.Material.Names:
                    mat = obj.Material.Materials[obj.Material.Names.index("Solid panel")]
                    if ("DiffuseColor" in mat.Material) and ("(" in mat.Material["DiffuseColor"]):
                        panelcolor = tuple(
                            [float(f) for f in mat.Material["DiffuseColor"].strip("()").split(",")]
                        )
                    paneltransparency = 0
        basecolor = basecolor[:3] + (1.0 - basetransparency,)
        panelcolor = panelcolor[:3] + (1.0 - paneltransparency,)
        colors = []
        nmullions = (
            obj.VerticalMullionNumber + obj.HorizontalMullionNumber + obj.DiagonalMullionNumber
        )
        for i, solid in enumerate(obj.Shape.Solids):
            for _ in solid.Faces:
                if i < nmullions:
                    colors.append(basecolor)
                else:
                    colors.append(panelcolor)
        if self.areDifferentColors(colors, obj.ViewObject.DiffuseColor) or force:
            obj.ViewObject.DiffuseColor = colors
